"""PCK command parsers and generators."""

from __future__ import annotations

import re
from collections.abc import Sequence

from pypck import lcn_defs
from pypck.lcn_addr import LcnAddr


class PckParser:
    """Helpers to parse LCN-PCK commands.

    LCN-PCK is the command-syntax used by LCN-PCHK to send and receive LCN
    commands.
    """

    # Authentication at LCN-PCHK: Request user name.
    AUTH_USERNAME = "Username:"

    # Authentication at LCN-PCHK: Request password.
    AUTH_PASSWORD = "Password:"

    # Authentication at LCN-PCHK succeeded.
    AUTH_OK = "OK"

    # Authentication at LCN-PCHK failed.
    AUTH_FAILED = "Authentification failed."

    # LCN-PK/PKU is connected.
    LCNCONNSTATE_CONNECTED = "$io:#LCN:connected"

    # LCN-PK/PKU is disconnected.
    LCNCONNSTATE_DISCONNECTED = "$io:#LCN:disconnected"

    # Decimal mode set
    DEC_MODE_SET = "(dec-mode)"

    # License Error
    LICENSE_ERROR = "$err:(license?)"

    # Pattern to parse error messages.
    PATTERN_COMMAND_ERROR = re.compile(r"\((?P<message>.+)\?\)")

    # Pattern to parse ping messages.
    PATTERN_PING = re.compile(r"\^ping(?P<count>\d*)-")

    # Pattern to parse positive acknowledges.
    PATTERN_ACK_POS = re.compile(r"-M(?P<seg_id>\d{3})(?P<mod_id>\d{3})!")

    # Pattern to parse negative acknowledges.
    PATTERN_ACK_NEG = re.compile(r"-M(?P<seg_id>\d{3})(?P<mod_id>\d{3})(?P<code>\d+)")

    # Pattern to parse segment coupler responses.
    PATTERN_SK_RESPONSE = re.compile(
        r"=M(?P<seg_id>\d{3})(?P<mod_id>\d{3})\.SK(?P<id>\d+)"
    )

    # Pattern to parse serial number and firmware date responses.
    PATTERN_SN = re.compile(
        r"=M(?P<seg_id>\d{3})(?P<mod_id>\d{3})\.SN(?P<hardware_serial>[0-9|A-F]{10})"
        r"(?P<manu>.{2})FW(?P<software_serial>[0-9|A-F]{6})"
        r"HW(?P<hardware_type>\d+)"
    )

    # Pattern to parse module name and comment
    PATTERN_NAME_COMMENT = re.compile(
        r"=M(?P<seg_id>\d{3})(?P<mod_id>\d{3})\.(?P<command>[NKO])"
        r"(?P<block_id>\d)(?P<text>.{0,12})"
    )

    # Pattern to parse the static and dynamic group membership status
    PATTERN_STATUS_GROUPS = re.compile(
        r"=M(?P<seg_id>\d{3})(?P<mod_id>\d{3})\.G(?P<kind>[DP])(?P<max_groups>\d{3})"
        r"(?:(?P<g1>\d{3}))?(?:(?P<g2>\d{3}))?(?:(?P<g3>\d{3}))?(?:(?P<g4>\d{3}))?"
        r"(?:(?P<g5>\d{3}))?(?:(?P<g6>\d{3}))?(?:(?P<g7>\d{3}))?(?:(?P<g8>\d{3}))?"
        r"(?:(?P<g9>\d{3}))?(?:(?P<g10>\d{3}))?(?:(?P<g11>\d{3}))?(?:(?P<g12>\d{3}))?"
    )

    # Pattern to parse output-port status responses in percent.
    PATTERN_STATUS_OUTPUT_PERCENT = re.compile(
        r":M(?P<seg_id>\d{3})(?P<mod_id>\d{3})A(?P<output_id>\d)(?P<percent>\d+)"
    )

    # Pattern to parse output-port status responses in native format (0..200).
    PATTERN_STATUS_OUTPUT_NATIVE = re.compile(
        r":M(?P<seg_id>\d{3})(?P<mod_id>\d{3})O(?P<output_id>\d)(?P<value>\d+)"
    )

    # Pattern to parse relays status responses.
    PATTERN_STATUS_RELAYS = re.compile(
        r":M(?P<seg_id>\d{3})(?P<mod_id>\d{3})Rx(?P<byte_value>\d+)"
    )

    # Pattern to parse binary-sensors status responses.
    PATTERN_STATUS_BINSENSORS = re.compile(
        r":M(?P<seg_id>\d{3})(?P<mod_id>\d{3})Bx(?P<byte_value>\d+)"
    )

    # Pattern to parse variable 1-12 status responses (since 170206).
    PATTERN_STATUS_VAR = re.compile(
        r"%M(?P<seg_id>\d{3})(?P<mod_id>\d{3})\.A(?P<id>\d{3})(?P<value>\d+)"
    )

    # Pattern to parse set-point variable status responses (since 170206).
    PATTERN_STATUS_SETVAR = re.compile(
        r"%M(?P<seg_id>\d{3})(?P<mod_id>\d{3})\.S(?P<id>\d)(?P<value>\d+)"
    )

    # Pattern to parse threshold status responses (since 170206).
    PATTERN_STATUS_THRS = re.compile(
        r"%M(?P<seg_id>\d{3})(?P<mod_id>\d{3})\.T(?P<register_id>\d)"
        r"(?P<thrs_id>\d)(?P<value>\d+)"
    )

    # Pattern to parse S0-input status responses (LCN-BU4L).
    PATTERN_STATUS_S0INPUT = re.compile(
        r"%M(?P<seg_id>\d{3})(?P<mod_id>\d{3})\.C(?P<id>\d)(?P<value>\d+)"
    )

    # Pattern to parse generic variable status responses (concrete type
    # unknown, before 170206).
    PATTERN_VAR_GENERIC = re.compile(
        r"%M(?P<seg_id>\d{3})(?P<mod_id>\d{3})\.(?P<value>\d+)"
    )

    # Pattern to parse threshold register 1 status responses (5 values,
    # before 170206). */
    PATTERN_THRS5 = re.compile(
        r"=M(?P<seg_id>\d{3})(?P<mod_id>\d{3})\.S1(?P<value1>\d{5})"
        r"(?P<value2>\d{5})(?P<value3>\d{5})(?P<value4>\d{5})"
        r"(?P<value5>\d{5})(?P<hyst>\d{5})"
    )

    # Pattern to parse status of LEDs and logic-operations responses.
    PATTERN_STATUS_LEDSANDLOGICOPS = re.compile(
        r"=M(?P<seg_id>\d{3})(?P<mod_id>\d{3})\.TL(?P<led_states>[AEBF]{12})"
        r"(?P<logic_op_states>[NTV]{4})"
    )

    # Pattern to parse key-locks status responses.
    PATTERN_STATUS_KEYLOCKS = re.compile(
        r"=M(?P<seg_id>\d{3})(?P<mod_id>\d{3})\.TX(?P<table0>\d{3})"
        r"(?P<table1>\d{3})(?P<table2>\d{3})((?P<table3>\d{3}))?"
    )

    # Pattern to parse scene output status messages.
    PATTERN_STATUS_SCENE_OUTPUTS = re.compile(
        r"=M(?P<seg_id>\d{3})(?P<mod_id>\d{3})\.SZ(?P<scene_id>\d{3})"
        r"(?P<output1>\d{3})(?P<ramp1>\d{3})(?P<output2>\d{3})(?P<ramp2>\d{3})"
        r"(?P<output3>\d{3})(?P<ramp3>\d{3})(?P<output4>\d{3})(?P<ramp4>\d{3})"
    )

    # Pattern to parse send command to host messages.
    PATTERN_SEND_COMMAND_HOST = re.compile(
        r"(\$M|\+M004)(?P<seg_id>\d{3})(?P<mod_id>\d{3})\.SKH"
        r"(?P<p1>\d{3})(?P<p2>\d{3})"
        r"(?:(?P<p3>\d{3})(?P<p4>\d{3})(?P<p5>\d{3})(?P<p6>\d{3}))?"
        r"(?:(?P<p7>\d{3})(?P<p8>\d{3})(?P<p9>\d{3})(?P<p10>\d{3})"
        r"(?P<p11>\d{3})(?P<p12>\d{3})(?P<p13>\d{3})(?P<p14>\d{3}))?"
    )

    # Pattern to parse send key to host messages.
    PATTERN_SEND_KEYS_HOST = re.compile(
        r"(\$M|\+M004)(?P<seg_id>\d{3})(?P<mod_id>\d{3})\.STH"
        r"(?P<actions>\d{3})(?P<keys>\d{3})"
    )

    # Pattern to parse transmitter status messages.
    PATTERN_STATUS_TRANSMITTER = re.compile(
        r"=M(?P<seg_id>\d{3})(?P<mod_id>\d{3})\.ZI"
        r"(?P<code1>\d{3})(?P<code2>\d{3})(?P<code3>\d{3})"
        r"(?P<level>\d{2})(?P<key>\d)(?P<action>\d{3})"
    )

    # Pattern to parse transponder status messages.
    PATTERN_STATUS_TRANSPONDER = re.compile(
        r"=M(?P<seg_id>\d{3})(?P<mod_id>\d{3})\.ZT"
        r"(?P<code1>\d{3})(?P<code2>\d{3})(?P<code3>\d{3})"
    )

    # Pattern to parse fingerprint status messages.
    PATTERN_STATUS_FINGERPRINT = re.compile(
        r"=M(?P<seg_id>\d{3})(?P<mod_id>\d{3})\.ZF"
        r"(?P<code1>\d{3})(?P<code2>\d{3})(?P<code3>\d{3})"
    )

    # Pattern to parse codelock status messages.
    PATTERN_STATUS_CODELOCK = re.compile(
        r"=M(?P<seg_id>\d{3})(?P<mod_id>\d{3})\.ZC"
        r"(?P<code1>\d{3})(?P<code2>\d{3})(?P<code3>\d{3})"
    )

    # Pattern to parse motor position BS4 status messages.
    PATTERN_STATUS_MOTOR_POSITION_BS4 = re.compile(
        r"=M(?P<seg_id>\d{3})(?P<mod_id>\d{3})\."
        r"RM(?P<motor1_id>[1-4])(?P<position1>[0-9]{3})(?P<limit1>[0-9]{3}|\?)"
        r"(?P<time_down1>[0-9]{5}|\?)(?P<time_up1>[0-9]{5}|\?)"
        r"RM(?P<motor2_id>[1-4])(?P<position2>[0-9]{3})(?P<limit2>[0-9]{3}|\?)"
        r"(?P<time_down2>[0-9]{5}|\?)(?P<time_up2>[0-9]{5}|\?)"
    )

    # Pattern to parse motor position module status messages.
    PATTERN_STATUS_MOTOR_POSITION_MODULE = re.compile(
        r":M(?P<seg_id>\d{3})(?P<mod_id>\d{3})"
        r"P(?P<motor_id>[1-4])(?P<position>[0-9]{3})"
    )

    @staticmethod
    def get_boolean_value(input_byte: int) -> list[bool]:
        """Get boolean representation for the given byte.

        :param    int    input_byte:    Input byte as int8.

        :return:    List with 8 boolean values.
        :rtype:     list
        """
        if (input_byte < 0) or (input_byte > 255):
            raise ValueError("Invalid input_byte.")

        result = []
        for i in range(8):
            result.append((input_byte & (1 << i)) != 0)
        return result


class PckGenerator:
    """Helpers to generate LCN-PCK commands.

    LCN-PCK is the command-syntax used by LCN-PCHK to send and receive LCN
    commands.
    """

    # Terminates a PCK command
    TERMINATION = "\n"

    TABLE_NAMES = ["A", "B", "C", "D"]

    @staticmethod
    def ping(counter: int) -> str:
        """Generate a keep-alive.

        LCN-PCHK will close the connection if it does not receive any commands
        specific period (10 minutes by default).

        :param    int    counter:    The current ping's id (optional, but
                                     'best practice'). Should start with 1
        :return:    The PCK command as text
        :rtype:    str
        """
        return f"^ping{counter:d}"

    @staticmethod
    def set_dec_mode() -> str:
        """Generate PCK command to set used number system to decimal."""
        return "!CHD"

    @staticmethod
    def set_operation_mode(
        dim_mode: lcn_defs.OutputPortDimMode, status_mode: lcn_defs.OutputPortStatusMode
    ) -> str:
        """Generate a PCK command to set the connection's operation mode.

        This influences how output-port commands and status are interpreted
        and must be in sync with the LCN bus.

        :param    OuputPortDimMode        dimMode:        The dimming mode
                                                          (50/200 steps)
        :param    OutputPortStatusMode    statusMode:     The status mode
                                                          (percent/native)
        :return:    The PCK command as text
        :rtype:    str
        """
        return (
            "!OM"
            f"{'1' if (dim_mode == lcn_defs.OutputPortDimMode.STEPS200) else '0'}"
            f"{'P' if (status_mode == lcn_defs.OutputPortStatusMode.PERCENT) else 'N'}"
        )

    @staticmethod
    def generate_address_header(
        addr: LcnAddr, local_seg_id: int, wants_ack: bool
    ) -> str:
        """Generate a PCK command address header.

        :param    addr:   The module's/group's address
        :type     addr:   :class:`~LcnAddr`
        :param    int     local_seg_id:    The local segment id
        :param    bool    wants_ack:      Is an acknowledge requested.

        :return:    The PCK address header string.
        :rtype:     str
        """
        return (
            ">"
            f"{'G' if addr.is_group else 'M'}"
            f"{addr.get_physical_seg_id(local_seg_id):03d}"
            f"{addr.addr_id:03d}"
            f"{'!' if wants_ack else '.'}"
        )

    @staticmethod
    def segment_coupler_scan() -> str:
        """Generate a scan-command for LCN segment-couplers.

        Used to detect the local segment (where the physical bus connection is
        located).

        :return:    The PCK command (without address header) as text
        :rtype:    str
        """
        return "SK"

    @staticmethod
    def request_serial() -> str:
        """Generate a firmware/serial-number request.

        :return: The PCK command (without address header) as text
        :rtype:    str
        """
        return "SN"

    @staticmethod
    def request_name(block_id: int) -> str:
        """Generate a name request.

        :return The PCK command (without address header) as text
        :rtype:    str
        """
        if block_id not in [0, 1]:
            raise ValueError("Invalid block_id.")
        return f"NMN{block_id + 1}"

    @staticmethod
    def request_comment(block_id: int) -> str:
        """Generate a comment request.

        :return The PCK command (without address header) as text
        :rtype:    str
        """
        if block_id not in [0, 1, 2]:
            raise ValueError("Invalid block_id.")
        return f"NMK{block_id + 1}"

    @staticmethod
    def request_oem_text(block_id: int) -> str:
        """Generate an oem text request.

        :return The PCK command (without address header) as text
        :rtype:    str
        """
        if block_id not in [0, 1, 2, 3]:
            raise ValueError("Invalid block_id.")
        return f"NMO{block_id + 1}"

    @staticmethod
    def request_group_membership_static() -> str:
        """Generate a group membership request for static membership (EEPROM).

        :return The PCK command (without address header) as text
        :rtype:    str
        """
        return "GP"

    @staticmethod
    def request_group_membership_dynamic() -> str:
        """Generate a group membership request for dynamic membership.

        :return The PCK command (without address header) as text
        :rtype:    str
        """
        return "GD"

    @staticmethod
    def request_output_status(output_id: int) -> str:
        """Generate an output-port status request.

        :param    int    output_id:    Output id 0..3
        :return:    The PCK command (without address header) as text
        :rtype:    str
        """
        if (output_id < 0) or (output_id > 3):
            raise ValueError("Invalid output_id.")
        return f"SMA{output_id + 1}"

    @staticmethod
    def dim_output(output_id: int, percent: float, ramp: int) -> str:
        """Generate a dim command for a single output-port.

        :param    int    output_id:    Output id 0..3
        :param    float  percent:      Brightness in percent 0..100
        :param    int    ramp:         Ramp value
        :return:    The PCK command (without address header) as text
        :rtype:    str
        """
        if (output_id < 0) or (output_id > 3):
            raise ValueError("Invalid output_id.")
        percent_round = int(round(percent * 2))
        ramp = int(ramp)
        if (percent_round % 2) == 0:
            # Use the percent command (supported by all LCN-PCHK versions)
            pck = f"A{output_id + 1}DI{percent_round // 2:03d}{ramp:03d}"
        else:
            # We have a ".5" value. Use the native command (supported since
            # LCN-PCHK 2.3)
            pck = f"O{output_id + 1}DI{percent_round:03d}{ramp:03d}"
        return pck

    @staticmethod
    def dim_all_outputs(percent: float, ramp: int, software_serial: int) -> str:
        """Generate a dim command for all output-ports.

        :param    float  percent:           Brightness in percent 0..100
        :param    int    ramp:              Ramp value
        :param    int    software_serial:   The expected firmware version of all
                                            receiving modules.
        :return:    The PCK command (without address header) as text
        :rtype:    str
        """
        percent_round = int(round(percent * 2))
        if software_serial >= 0x180501:
            # Supported since LCN-PCHK 2.61
            pck = (
                "OY"
                f"{percent_round:03d}"
                f"{percent_round:03d}"
                f"{percent_round:03d}"
                f"{percent_round:03d}"
                f"{ramp:03d}"
            )
        elif percent_round == 0:  # All off
            pck = f"AA{ramp:03d}"
        elif percent_round == 200:  # All on
            pck = f"AE{ramp:03d}"
        else:
            # This is our worst-case: No high-res, no ramp
            pck = f"AH{percent_round // 2:03d}"
        return pck

    @staticmethod
    def rel_output(output_id: int, percent: float) -> str:
        """Generate a command to change the value of an output-port.

        :param    int    output_id:    Output id 0..3
        :param    float  percent:      Relative percentage -100..100
        :return:    The PCK command (without address header) as text
        :rtype:    str
        """
        if (output_id < 0) or (output_id > 3):
            raise ValueError("Invalid output_id.")

        percent_round = int(round(percent * 2))
        if percent_round % 2 == 0:
            # Use the percent command (supported by all LCN-PCHK versions)
            pck = (
                "A"
                f"{output_id + 1}"
                f"{'AD' if percent >= 0 else 'SB'}"
                f"{abs(percent_round // 2):03d}"
            )
        else:
            # We have a ".5" value. Use the native command (supported since
            # LCN-PCHK 2.3)
            pck = (
                "O"
                f"{output_id + 1}"
                f"{'AD' if percent >= 0 else 'SB'}"
                f"{abs(percent_round):03d}"
            )
        return pck

    @staticmethod
    def toggle_output(output_id: int, ramp: int, to_memory: bool = False) -> str:
        """Generate a command that toggles a single output-port.

        Toggle mode: (on->off, off->on).

        :param    int    output_id:    Output id 0..3
        :param    int    ramp:         Ramp value
        :param    bool   to_memory:    If True, the dimming status is stored

        :return:    The PCK command (without address header) as text
        :rtype:    str
        """
        if (output_id < 0) or (output_id > 3):
            raise ValueError("Invalid output_id.")
        return f"A{output_id + 1}{'MT' if to_memory else 'TA'}{ramp:03d}"

    @staticmethod
    def toggle_all_outputs(ramp: int) -> str:
        """Generate a command that toggles all output-ports.

        Toggle mode: (on->off, off->on).

        :param    int    ramp:    Ramp value
        :return:    The PCK command (without address header) as text
        :rtype:    str
        """
        return f"AU{ramp:03d}"

    @staticmethod
    def request_relays_status() -> str:
        """Generate a relays-status request.

        :return: The PCK command (without address header) as text
        :rtype:    str
        """
        return "SMR"

    @staticmethod
    def control_relays(states: list[lcn_defs.RelayStateModifier]) -> str:
        """Generate a command to control relays.

        :param     RelayStateModifier    states:    The 8 modifiers for the
                                                    relay states as a list
        :return:   The PCK command (without address header) as text
        :rtype:    str
        """
        if len(states) != 8:
            raise ValueError("Invalid states length.")
        ret = "R8"
        for state in states:
            ret += state.value

        return ret

    @staticmethod
    def control_relays_timer(
        time_msec: int, states: list[lcn_defs.RelayStateModifier]
    ) -> str:
        """Generate a command to control relays.

        :param     int                   time_msec: Duration of timer in
                                                    milliseconds
        :param     RelayStateModifier    states:    The 8 modifiers for the
                                                    relay states as a list
                                                    (only ON and OFF allowed)
        :return:   The PCK command (without address header) as text
        :rtype:    str
        """
        if len(states) != 8:
            raise ValueError("Invalid states length.")
        value = lcn_defs.time_to_native_value(time_msec)
        ret = f"R8T{value:03d}"
        for state in states:
            assert state in (
                lcn_defs.RelayStateModifier.ON,
                lcn_defs.RelayStateModifier.NOCHANGE,
            )
            ret += state.value

        return ret

    @staticmethod
    def control_motor_relays(
        motor_id: int,
        state: lcn_defs.MotorStateModifier,
        mode: lcn_defs.MotorPositioningMode = lcn_defs.MotorPositioningMode.NONE,
    ) -> str:
        """Generate a command to control motors via relays.

        :param    int                 motor_id:    The motor id 0..3
        :param    MotorStateModifier  state:       The modifier for the
                                                   motor state
        :return:  The PCK command (without address header) as text
        :rtype:   str
        """
        if 0 > motor_id > 3:
            raise ValueError("Invalid motor id")

        if mode not in lcn_defs.MotorPositioningMode:
            raise ValueError("Wrong motor position mode")

        if mode == lcn_defs.MotorPositioningMode.BS4:
            new_motor_id = [1, 2, 5, 6][motor_id]
            if state == lcn_defs.MotorStateModifier.DOWN:
                # AU=window open / cover down
                action = "AU"
            elif state == lcn_defs.MotorStateModifier.UP:
                # ZU=window close / cover up
                action = "ZU"
            elif state == lcn_defs.MotorStateModifier.STOP:
                action = "ST"
            else:
                raise ValueError("Invalid motor state for BS4 mode")

            return f"R8M{new_motor_id}{action}"

        # lcn_defs.MotorPositioningMode.NONE
        # lcn_defs.MotorPositioningMode.MODULE
        if state == lcn_defs.MotorStateModifier.UP:
            port_onoff = lcn_defs.RelayStateModifier.ON
            port_updown = lcn_defs.RelayStateModifier.OFF
        elif state == lcn_defs.MotorStateModifier.DOWN:
            port_onoff = lcn_defs.RelayStateModifier.ON
            port_updown = lcn_defs.RelayStateModifier.ON
        elif state == lcn_defs.MotorStateModifier.STOP:
            port_onoff = lcn_defs.RelayStateModifier.OFF
            port_updown = lcn_defs.RelayStateModifier.NOCHANGE
        elif state == lcn_defs.MotorStateModifier.TOGGLEONOFF:
            port_onoff = lcn_defs.RelayStateModifier.TOGGLE
            port_updown = lcn_defs.RelayStateModifier.NOCHANGE
        elif state == lcn_defs.MotorStateModifier.TOGGLEDIR:
            port_onoff = lcn_defs.RelayStateModifier.NOCHANGE
            port_updown = lcn_defs.RelayStateModifier.TOGGLE
        elif state == lcn_defs.MotorStateModifier.CYCLE:
            port_onoff = lcn_defs.RelayStateModifier.TOGGLE
            port_updown = lcn_defs.RelayStateModifier.TOGGLE
        elif state == lcn_defs.MotorStateModifier.NOCHANGE:
            port_onoff = lcn_defs.RelayStateModifier.NOCHANGE
            port_updown = lcn_defs.RelayStateModifier.NOCHANGE
        else:
            raise ValueError("Invalid motor state")

        states = [lcn_defs.RelayStateModifier.NOCHANGE] * 8
        states[motor_id * 2] = port_onoff
        states[motor_id * 2 + 1] = port_updown
        return "R8" + "".join([state.value for state in states])

    @staticmethod
    def control_motor_relays_position(
        motor_id: int, position: float, mode: lcn_defs.MotorPositioningMode
    ) -> str:
        """Control motor position via relays and BS4 or module.

        :param    int                   motor_id:   The motor port of the LCN module
        :param    float                 position:   The position to set in percentage (0..100)
                                                    (0: closed cover, 100: open cover)
        :param    MotorPositioningMode  mode:       The motor positioning mode

        :return:  The PCK command (without address header) as text
        :rtype:   str
        """
        if 0 > motor_id > 3:
            raise ValueError("Invalid motor")

        if mode == lcn_defs.MotorPositioningMode.BS4:
            new_motor_id = [1, 2, 5, 6][motor_id]
            action = f"GP{int(200 - 2 * position):03d}"
            return f"R8M{new_motor_id}{action}"
        elif mode == lcn_defs.MotorPositioningMode.MODULE:
            new_motor_id = 1 << motor_id
            return f"JH{100 - position:03d}{new_motor_id:03d}"
        else:
            raise ValueError("Wrong motor positioning mode")

    @staticmethod
    def request_motor_position_status(motor_pair: int) -> str:
        """Generate a motor position status request for BS4.

        :param    int    motor_pair:    Motor pair 0: 1, 2; 1: 3, 4

        :return:    The PCK command (without address header) as text
        :rtype:    str
        """
        if motor_pair not in [0, 1]:
            raise ValueError("Invalid motor_pair.")
        return f"R8M{7 if motor_pair else 3}P{motor_pair + 1}"

    @staticmethod
    def control_motor_outputs(
        state: lcn_defs.MotorStateModifier,
        reverse_time: lcn_defs.MotorReverseTime | None = None,
    ) -> str:
        """Generate a command to control a motor via output ports 1+2.

        :param    MotorStateModifier    state:     The modifier for the
                                                   motor state
        :param    MotorReverseTime      reverse_time: Reverse time for modules
                                                      with FW<190C
        :return:  The PCK command (without address header) as text
        :rtype:   str
        """
        if state == lcn_defs.MotorStateModifier.UP:
            if reverse_time in [None, lcn_defs.MotorReverseTime.RT70]:
                params = (0x01, 0xE4, 0x00)
                ret = f"X2{params[0]:03d}{params[1]:03d}{params[2]:03d}"
            elif reverse_time == lcn_defs.MotorReverseTime.RT600:
                ret = PckGenerator.dim_output(0, 100, 8)
            elif reverse_time == lcn_defs.MotorReverseTime.RT1200:
                ret = PckGenerator.dim_output(0, 100, 11)
            else:
                raise ValueError("Wrong MotorReverseTime.")

        elif state == lcn_defs.MotorStateModifier.DOWN:
            if reverse_time in [None, lcn_defs.MotorReverseTime.RT70]:
                params = (0x01, 0x00, 0xE4)
                ret = f"X2{params[0]:03d}{params[1]:03d}{params[2]:03d}"
            elif reverse_time == lcn_defs.MotorReverseTime.RT600:
                ret = PckGenerator.dim_output(1, 100, 8)
            elif reverse_time == lcn_defs.MotorReverseTime.RT1200:
                ret = PckGenerator.dim_output(1, 100, 11)
            else:
                raise ValueError("Wrong MotorReverseTime.")

        elif state == lcn_defs.MotorStateModifier.STOP:
            ret = "AY000000"
        elif state == lcn_defs.MotorStateModifier.CYCLE:
            ret = "JE"
        else:
            raise ValueError("MotorStateModifier is not supported by output ports.")

        return ret

    @staticmethod
    def request_bin_sensors_status() -> str:
        """Generate a binary-sensors status request.

        :return:    The PCK command (without address header) as text
        :rtype:    str
        """
        return "SMB"

    @staticmethod
    def var_abs(var: lcn_defs.Var, value: int) -> str:
        """Generate a command that sets a variable absolute.

        :param    Var    var:    The target variable to set
        :param    int    value:  The absolute value to set
        :return:    The PCK command (without address header) as text
        :rtype:    str
        """
        set_point_id = lcn_defs.Var.to_set_point_id(var)
        if set_point_id != -1:
            # Set absolute (not in PCK yet)
            byte1 = set_point_id << 6  # 01000000
            byte1 |= 0x20  # xx10xxxx (set absolute)
            value -= 1000  # Offset
            byte1 |= (value >> 8) & 0x0F  # xxxx1111
            byte2 = value & 0xFF
            return f"X2{30:03d}{byte1:03d}{byte2:03d}"

        # Setting variables and thresholds absolute not implemented in LCN
        # firmware yet
        raise ValueError("Wrong variable type.")

    @staticmethod
    def update_status_var(var: lcn_defs.Var, value: int) -> str:
        """Generate a command that send variable status updates.

        PCHK provides this variables by itself on selected segments
        is only possible with group 4

        :param    Var    var:    The target variable to set
        :param    int    value:  The absolute value to set
        :return:    The PCK command (without address header) as text
        :rtype:    str
        """
        var_id = lcn_defs.Var.to_var_id(var)
        if var_id != -1:
            # define variable to set, offset 0x01000000
            x2cmd = var_id | 0x40
            byte1 = (value >> 8) & 0xFF
            byte2 = value & 0xFF
            return f"X2{x2cmd:03d}{byte1:03d}{byte2:03d}"

        # Setting variables and thresholds absolute not implemented in LCN
        # firmware yet
        raise ValueError("Wrong variable type.")

    @staticmethod
    def var_reset(var: lcn_defs.Var, software_serial: int) -> str:
        """Generate a command that resets a variable to 0.

        :param    Var    var:               The target variable to set 0
        :param    int    software_serial:   The expected firmware version of all
                                            receiving modules.
        :return:    The PCK command (without address header) as text
        :rtype:    str
        """
        var_id = lcn_defs.Var.to_var_id(var)
        if var_id != -1:
            if software_serial >= 0x170206:
                pck = f"Z-{var_id + 1:03d}{4090:04d}"
            elif var_id == 0:
                pck = "ZS30000"
            else:
                raise ValueError("Wrong variable type.")
            return pck

        set_point_id = lcn_defs.Var.to_set_point_id(var)
        if set_point_id != -1:
            # Set absolute = 0 (not in PCK yet)
            byte1 = set_point_id << 6  # 01000000
            byte1 |= 0x20  # xx10xxxx 9set absolute)
            byte2 = 0
            return f"X2{30:03d}{byte1:03d}{byte2:03d}"

        # Reset for threshold not implemented in LCN firmware yet
        raise ValueError("Wrong variable type.")

    @staticmethod
    def var_rel(
        var: lcn_defs.Var,
        rel_var_ref: lcn_defs.RelVarRef,
        value: int,
        software_serial: int,
    ) -> str:
        """Generate a command to change the value of a variable.

        :param    Var       var:                The target variable to change
        :param    RelVarRef rel_var_ref:        The reference-point
        :param    int       value:              The native LCN value to add/subtract
                                                (can be negative)
        :param    int       software_serial:    The expected firmware version of all
                                                receiving modules.
        :return:    The PCK command (without address header) as text
        :rtype:    str
        """
        var_id = lcn_defs.Var.to_var_id(var)
        if var_id != -1:
            if var_id == 0:
                # Old command for variable 1 / T-var (compatible with all
                # modules)
                pck = f"Z{'A' if value >= 0 else 'S'}{abs(value)}"
            else:
                # New command for variable 1-12 (compatible with all modules,
                # since LCN-PCHK 2.8)
                pck = f"Z{'+' if value >= 0 else '-'}{var_id + 1:03d}{abs(value)}"
            return pck

        set_point_id = lcn_defs.Var.to_set_point_id(var)
        if set_point_id != -1:
            pck = (
                "RE"
                f"{'A' if set_point_id == 0 else 'B'}"
                f"S{'A' if rel_var_ref == lcn_defs.RelVarRef.CURRENT else 'P'}"
                f"{'+' if value >= 0 else '-'}"
                f"{abs(value)}"
            )
            return pck

        thrs_register_id = lcn_defs.Var.to_thrs_register_id(var)
        thrs_id = lcn_defs.Var.to_thrs_id(var)
        if (thrs_register_id != -1) and (thrs_id != -1):
            if software_serial >= 0x170206:
                # New command for registers 1-4 (since 170206, LCN-PCHK 2.8)
                pck = (
                    "SS"
                    f"{'R' if rel_var_ref == lcn_defs.RelVarRef.CURRENT else 'E'}"
                    f"{abs(value):04d}"
                    f"{'A' if value >= 0 else 'S'}"
                    f"R{thrs_register_id + 1}"
                    f"{thrs_id + 1}"
                )
            elif thrs_register_id == 0:
                # Old command for register 1 (before 170206)
                pck = (
                    "SS"
                    f"{'R' if rel_var_ref == lcn_defs.RelVarRef.CURRENT else 'E'}"
                    f"{abs(value):04d}"
                    f"{'A' if value >= 0 else 'S'}"
                    f"{'1' if thrs_id == 0 else '0'}"
                    f"{'1' if thrs_id == 1 else '0'}"
                    f"{'1' if thrs_id == 2 else '0'}"
                    f"{'1' if thrs_id == 3 else '0'}"
                    f"{'1' if thrs_id == 4 else '0'}"
                )
            return pck

        raise ValueError("Wrong variable type.")

    @staticmethod
    def request_var_status(var: lcn_defs.Var, software_serial: int) -> str:
        """Generate a variable value request.

        :param    Var    var:               The variable to request
        :param    int    software_serial:   The target module's firmware version
        :return:  The PCK command (without address header) as text
        :rtype:   str
        """
        if software_serial >= 0x170206:
            var_id = lcn_defs.Var.to_var_id(var)
            if var_id != -1:
                return f"MWT{var_id + 1}"

            set_point_id = lcn_defs.Var.to_set_point_id(var)
            if set_point_id != -1:
                return f"MWS{set_point_id + 1}"

            thrs_register_id = lcn_defs.Var.to_thrs_register_id(var)
            if thrs_register_id != -1:
                # Whole register
                return f"SE{thrs_register_id + 1}"

            s0_id = lcn_defs.Var.to_s0_id(var)
            if s0_id != -1:
                return f"MWC{s0_id + 1}"
        else:
            if var == lcn_defs.Var.VAR1ORTVAR:
                pck = "MWV"
            elif var == lcn_defs.Var.VAR2ORR1VAR:
                pck = "MWTA"
            elif var == lcn_defs.Var.VAR3ORR2VAR:
                pck = "MWTB"
            elif var == lcn_defs.Var.R1VARSETPOINT:
                pck = "MWSA"
            elif var == lcn_defs.Var.R2VARSETPOINT:
                pck = "MWSB"
            elif var in [
                lcn_defs.Var.THRS1,
                lcn_defs.Var.THRS2,
                lcn_defs.Var.THRS3,
                lcn_defs.Var.THRS4,
                lcn_defs.Var.THRS5,
            ]:
                pck = "SL1"  # Whole register
            return pck

        raise ValueError("Wrong variable type.")

    @staticmethod
    def request_leds_and_logic_ops() -> str:
        """Generate a request for LED and logic-operations states.

        :return: The PCK command (without address header) as text
        :rtype:  str
        """
        return "SMT"

    @staticmethod
    def control_led(led_id: int, state: lcn_defs.LedStatus) -> str:
        """Generate a command to the set the state of a single LED.

        :param    int          led_id:   Led id 0..11
        :param    LedStatus    state:    The state to set
        :return: The PCK command (without address header) as text
        :rtype:  str
        """
        if (led_id < 0) or (led_id > 11):
            raise ValueError("Bad led_id.")
        return f"LA{led_id + 1:03d}{state.value}"

    @staticmethod
    def send_keys(cmds: list[lcn_defs.SendKeyCommand], keys: list[bool]) -> str:
        """Generate a command to send LCN keys.

        :param    SendKeyCommand    cmds:    The 4 concrete commands to send
                                             for the tables (A-D) as list
        :param    list(bool)        keys:    The tables' 8 key-states (True
                                             means "send") as list

        :return:  The PCK command (without address header) as text
        :rtype:   str
        """
        if (len(cmds) != 4) or (len(keys) != 8):
            raise ValueError("Wrong cmds length or wrong keys length.")
        ret = "TS"
        for i, cmd in enumerate(cmds):
            if (cmd == lcn_defs.SendKeyCommand.DONTSEND) and (i == 3):
                # By skipping table D (if it is not used), we use the old
                # command
                # for table A-C which is compatible with older LCN modules
                break
            ret += cmd.value

        for key in keys:
            ret += "1" if key else "0"

        return ret

    @staticmethod
    def send_keys_hit_deferred(
        table_id: int, time: int, time_unit: lcn_defs.TimeUnit, keys: list[bool]
    ) -> str:
        """Generate a command to send LCN keys deferred / delayed.

        :param     int         table_id:    Table id 0(A)..3(D)
        :param     int         time:        The delay time
        :param     TimeUnit    time_unit:   The time unit
        :param     list(bool)  keys:        The table's 8 key-states (True
                                            means "send") as list
        :return:   The PCK command (without address header) as text
        :rtype:    str
        """
        if (table_id < 0) or (table_id > 3) or (len(keys) != 8):
            raise ValueError("Bad table_id or keys.")
        ret = "TV"
        try:
            ret += PckGenerator.TABLE_NAMES[table_id]
        except IndexError as exc:
            raise ValueError("Wrong table_id.") from exc

        ret += f"{time:03d}"
        if time_unit == lcn_defs.TimeUnit.SECONDS:
            if (time < 1) or (time > 60):
                raise ValueError("Wrong time.")
            ret += "S"
        elif time_unit == lcn_defs.TimeUnit.MINUTES:
            if (time < 1) or (time > 90):
                raise ValueError("Wrong time.")
            ret += "M"
        elif time_unit == lcn_defs.TimeUnit.HOURS:
            if (time < 1) or (time > 50):
                raise ValueError("Wrong time.")
            ret += "H"
        elif time_unit == lcn_defs.TimeUnit.DAYS:
            if (time < 1) or (time > 45):
                raise ValueError("Wrong time.")
            ret += "D"
        else:
            raise ValueError("Wrong time_unit.")

        for key in keys:
            ret += "1" if key else "0"

        return ret

    @staticmethod
    def request_key_lock_status() -> str:
        """Generate a request for key-lock states.

        Always requests table A-D. Supported since LCN-PCHK 2.8.

        :return:    The PCK command (without address header) as text
        :rtype:     str
        """
        return "STX"

    @staticmethod
    def lock_keys(table_id: int, states: list[lcn_defs.KeyLockStateModifier]) -> str:
        """Generate a command to lock keys.

        :param     int           table_id:  Table id 0(A)..3(D)
        :param     list(bool)    states:    The 8 key-lock modifiers as list
        :return:   The PCK command (without address header) as text
        :rtype:    str
        """
        if (table_id < 0) or (table_id > 3) or (len(states) != 8):
            raise ValueError("Bad table_id or states.")
        try:
            ret = f"TX{PckGenerator.TABLE_NAMES[table_id]}"
        except IndexError as exc:
            raise ValueError("Wrong table_id.") from exc

        for state in states:
            ret += state.value

        return ret

    @staticmethod
    def lock_keys_tab_a_temporary(
        time: int,
        time_unit: lcn_defs.TimeUnit,
        keys: list[lcn_defs.KeyLockStateModifier],
    ) -> str:
        """Generate a command to lock keys for table A temporary.

        There is no hardware-support for locking tables B-D.

        :param     int         time:         The lock time
        :param     TimeUnit    time_unit:    The time unit
        :param     list(bool)  keys:         The 8 key-lock states (True means
                                             lock) as list
        :return:   The PCK command (without address header) as text
        :rtype:    str
        """
        if len(keys) != 8:
            raise ValueError("Wrong keys length.")
        ret = f"TXZA{time:03d}"

        if time_unit == lcn_defs.TimeUnit.SECONDS:
            if (time < 1) or (time > 60):
                raise ValueError("Wrong time.")
            ret += "S"
        elif time_unit == lcn_defs.TimeUnit.MINUTES:
            if (time < 1) or (time > 90):
                raise ValueError("Wrong time.")
            ret += "M"
        elif time_unit == lcn_defs.TimeUnit.HOURS:
            if (time < 1) or (time > 50):
                raise ValueError("Wrong time.")
            ret += "H"
        elif time_unit == lcn_defs.TimeUnit.DAYS:
            if (time < 1) or (time > 45):
                raise ValueError("Wrong time.")
            ret += "D"
        else:
            raise ValueError("Wrong time_unit.")

        for key in keys:
            ret += "1" if key == lcn_defs.KeyLockStateModifier.ON else "0"

        return ret

    @staticmethod
    def dyn_text_part(row_id: int, part_id: int, part: bytes) -> bytes:
        """Generate the command header / start for sending dynamic texts.

        Used by LCN-GTxD periphery (supports 4 text rows).
        To complete the command, the text to send must be appended (UTF-8
        encoding).
        Texts are split up into up to 5 parts with 12 "UTF-8 bytes" each.

        :param    int    row_id:    Row id 0..3
        :param    int    part_id:   Part id 0..4
        :param    bytes  part:      Text part (up to 12 bytes), encoded as
                                    lcn_defs.LCN_ENCODING
        :return:  The PCK command (without address header) as encoded bytes
        :rtype:   bytes
        """
        if (
            (row_id < 0)
            or (row_id > 3)
            or (part_id < 0)
            or (part_id > 4)
            or (len(part) > 12)
        ):
            raise ValueError("Wrong row_id, part_id or part length.")
        return f"GTDT{row_id + 1}{part_id + 1}".encode() + part

    @staticmethod
    def lock_regulator(
        reg_id: int,
        state: bool,
        software_serial: int,
        target_value: float = -1,
    ) -> str:
        """Generate a command to lock a regulator.

        :param    int    reg_id:    Regulator id 0..1
        :param    bool   state:     The lock state
        :param    int    software_serial: The expected firmware version of all
                                           receiving modules.
        :param    foat    target_value: The target value in percent (use -1 to ignore)
        :return:  The PCK command (without address header) as text
        :rtype:   str
        """
        if (reg_id < 0) or (reg_id > 1):
            raise ValueError("Wrong reg_id.")
        if ((target_value < 0) or (target_value > 100)) and (target_value != -1):
            raise ValueError("Wrong target_value.")
        if (target_value != -1) and (software_serial >= 0x120301) and state:
            reg_byte = reg_id * 0x40 + 0x07
            return f"X2{0x1E:03d}{reg_byte:03d}{int(2 * target_value):03d}"
        return f"RE{'A' if reg_id == 0 else 'B'}X{'S' if state else 'A'}"

    @staticmethod
    def lock_thresholds(
        register_id: int,
        states: list[lcn_defs.ThresholdLockStateModifier],
    ) -> str:
        """Generate a command to lock thresholds of a register.

        Only supported for software serial >170206.
        :param    int    register_id:   Register id 0..3
        :param    list(ThresholdLockStateModifier) states: The 4 threshold lock modifiers as list
        :return:  The PCK command (without address header) as text
        :rtype:   str
        """
        if (register_id < 0) or (register_id > 3) or (len(states) != 4):
            raise ValueError("Wrong register_id or states length.")
        ret = f"SX{register_id + 1}"
        for state in states:
            ret += state.value

        return ret

    @staticmethod
    def change_scene_register(register_id: int) -> str:
        """Change the active scene register.

        :param    int    register_id:    Register id 0..9
        :return:  The PCK command (without address header) as text
        :rtype:   str
        """
        if (register_id < 0) or (register_id > 9):
            raise ValueError("Wrong register_id.")
        return f"SZW{register_id:03d}"

    @staticmethod
    def store_scene_outputs_direct(
        register_id: int, scene_id: int, percents: Sequence[float], ramps: Sequence[int]
    ) -> str:
        """Store the given output values and ramps in the given scene.

        :param    int           register_id: Register id 0..9
        :param    int           scene_id:    Scene id 0..9
        :param    list(float)   percents:    Output values in percent as list
        :param    list(int)     ramp:        Ramp values as list
        :return:  The PCK command (without address header) as text
        :rtype:   str
        """
        if (scene_id < 0) or (scene_id > 9):
            raise ValueError("Wrong scene_id.")
        if len(percents) not in (2, 4):
            raise ValueError("Need 2 or 4 output percent values.")
        if len(ramps) != len(percents):
            raise ValueError("Need as many ramp values as output percent values.")
        cmd = f"SZD{register_id:03d}{scene_id:03d}"
        for i, percent in enumerate(percents):
            cmd += f"{int(percent * 2):03d}{ramps[i]:03d}"
        return cmd

    @staticmethod
    def activate_scene_output(
        scene_id: int,
        output_ports: Sequence[lcn_defs.OutputPort] = (),
        ramp: int | None = None,
    ) -> str:
        """Activate the stored output states for the given scene.

        Please note: The output ports 3 and 4 can only be activated
        simultaneously. If one of them is given, the other one is activated,
        too.

        :param    int                scene_id:       Scene id 0..9
        :param    list(OutputPort)   output_ports:   Output ports to activate
                                                     as list
        :param    int                ramp:           Ramp value
        :return:  The PCK command (without address header) as text
        :rtype:   str
        """
        return PckGenerator._activate_or_store_scene_output(
            scene_id, output_ports, ramp, store=False
        )

    @staticmethod
    def store_scene_output(
        scene_id: int,
        output_ports: Sequence[lcn_defs.OutputPort] = (),
        ramp: int | None = None,
    ) -> str:
        """Store the current output states in the given scene.

        Please note: The output ports 3 and 4 can only be stored
        simultaneously. If one of them is given, the other one is stored,
        too.

        :param    int                scene_id:       Scene id 0..9
        :param    list(OutputPort)   output_ports:   Output ports to store
                                                     as list
        :param    int                ramp:           Ramp value
        :return:  The PCK command (without address header) as text
        :rtype:   str
        """
        return PckGenerator._activate_or_store_scene_output(
            scene_id, output_ports, ramp, store=True
        )

    @staticmethod
    def _activate_or_store_scene_output(
        scene_id: int,
        output_ports: Sequence[lcn_defs.OutputPort] = (),
        ramp: int | None = None,
        store: bool = False,
    ) -> str:
        if (scene_id < 0) or (scene_id > 9):
            raise ValueError("Wrong scene_id.")
        if not output_ports:
            raise ValueError("output_port list is empty.")
        output_mask = 0
        if lcn_defs.OutputPort.OUTPUT1 in output_ports:
            output_mask += 1
        if lcn_defs.OutputPort.OUTPUT2 in output_ports:
            output_mask += 2
        if (
            lcn_defs.OutputPort.OUTPUT3 in output_ports
            or lcn_defs.OutputPort.OUTPUT4 in output_ports
        ):
            output_mask += 4
        if store:
            action = "S"
        else:
            action = "A"
        if ramp is None:
            pck = f"SZ{action:s}{output_mask:1d}{scene_id:03d}"
        else:
            pck = f"SZ{action:s}{output_mask:1d}{scene_id:03d}{ramp:03d}"
        return pck

    @staticmethod
    def activate_scene_relay(
        scene_id: int, relay_ports: Sequence[lcn_defs.RelayPort] = ()
    ) -> str:
        """Activate the stored relay states for the given scene.

        :param    int                scene_id:       Scene id 0..9
        :param    list(RelayPort)    relay_ports:    Relay ports to activate
                                                     as list
        :return:  The PCK command (without address header) as text
        :rtype:   str
        """
        return PckGenerator._activate_or_store_scene_relay(
            scene_id, relay_ports, store=False
        )

    @staticmethod
    def store_scene_relay(
        scene_id: int, relay_ports: Sequence[lcn_defs.RelayPort] = ()
    ) -> str:
        """Store the current relay states in the given scene.

        :param    int                scene_id:       Scene id 0..9
        :param    list(RelayPort)    relay_ports:    Relay ports to store
                                                     as list
        :return:  The PCK command (without address header) as text
        :rtype:   str
        """
        return PckGenerator._activate_or_store_scene_relay(
            scene_id, relay_ports, store=True
        )

    @staticmethod
    def _activate_or_store_scene_relay(
        scene_id: int,
        relay_ports: Sequence[lcn_defs.RelayPort] = (),
        store: bool = False,
    ) -> str:
        if (scene_id < 0) or (scene_id > 9):
            raise ValueError("Wrong scene_id.")
        if not relay_ports:
            raise ValueError("relay_port list is empty.")
        relays_mask = ["0"] * 8
        for port in relay_ports:
            relays_mask[port.value] = "1"
        if store:
            action = "S"
        else:
            action = "A"
        return f"SZ{action}0{scene_id:03d}{''.join(relays_mask)}"

    @staticmethod
    def request_status_scene(register_id: int, scene_id: int) -> str:
        """Request the stored output and ramp values for the given scene.

        :param    int    register_id:    Register id 0..9
        :param    int    register_id:    Scene id 0..9
        :return:  The PCK command (without address header) as text
        :rtype:   str
        """
        if (register_id < 0) or (register_id > 9):
            raise ValueError("Wrong register_id.")
        if (scene_id < 0) or (scene_id > 9):
            raise ValueError("Wrong scene_id.")
        return f"SZR{register_id:03d}{scene_id:03d}"

    @staticmethod
    def beep(sound: lcn_defs.BeepSound, count: int) -> str:
        """Make count number of beep sounds.

        :param    BeepSound sound:  Beep sound style
        :param    int       count:  Number of beeps (1..15)
        :return:  The PCK command (without address header) as text
        :rtype:   str
        """
        if (count < 1) or (count > 15):
            raise ValueError("Wrong number of beeps.")
        return f"PI{sound.value:s}{count:03d}"

    @staticmethod
    def empty() -> str:
        """Generate an empty command (LEER) that does nothing.

        Combine with request for acknowledgement to discover and
        ping modules and to discover and verify group memberships.

        :return:  The PCK command (without address header) as text
        :rtype:   str
        """
        return "LEER"



"""Tests for variable value and unit handling."""

import math

import pytest

from pypck.lcn_defs import VarUnit, VarValue

VARIABLE_TEST_VALUES = (
    0,
    48,
    49,
    50,
    51,
    52,
    100,
    198,
    199,
    200,
    201,
    202,
    205,
    1000,
    1023,
    4095,
    65535,
)

UNITS_LOSSLESS_ROUNDTRIP = (
    VarUnit.NATIVE,
    VarUnit.CELSIUS,
    VarUnit.KELVIN,
    VarUnit.FAHRENHEIT,
    VarUnit.LUX_I,
    VarUnit.METERPERSECOND,
    VarUnit.PERCENT,
    VarUnit.VOLT,
    VarUnit.AMPERE,
    VarUnit.DEGREE,
    VarUnit.PPM,
    # VarUnit.LUX_T,
)

ROUNDTRIP_TEST_VECTORS = (
    *(
        (unit, value, value)
        for value in VARIABLE_TEST_VALUES
        for is_abs in (True, False)
        for unit in UNITS_LOSSLESS_ROUNDTRIP
    ),
)


CALIBRATION_TEST_VECTORS = (
    *(
        (VarUnit.CELSIUS, native, value)
        for native, value in ((0, -100), (1000, 0), (2000, 100))
    ),
    *(
        (VarUnit.KELVIN, native, value)
        for native, value in (
            (0, -100 + 273.15),
            (1000, 0 + 273.15),
            (2000, 100 + 273.15),
        )
    ),
    *(
        (VarUnit.FAHRENHEIT, native, value)
        for native, value in (
            (0, -100 * 1.8 + 32),
            (1000, 0 * 1.8 + 32),
            (2000, 100 * 1.8 + 32),
        )
    ),
    *(
        (VarUnit.LUX_I, native, value)
        for native, value in (
            (0, math.exp(0)),
            (10, math.exp(0.1)),
            (100, math.exp(1)),
            (1000, math.exp(10)),
        )
    ),
    *(
        (VarUnit.METERPERSECOND, native, value)
        for native, value in ((0, 0), (10, 1), (100, 10), (1000, 100))
    ),
    *(
        (VarUnit.PERCENT, native, value)
        for native, value in ((0, 0), (1, 1), (100, 100))
    ),
    *(
        (VarUnit.VOLT, native, value)
        for native, value in ((0, 0), (400, 1), (4000, 10))
    ),
    *(
        (VarUnit.AMPERE, native, value)
        for native, value in ((0, 0), (100, 0.001), (4000, 0.04))
    ),
    *(
        (VarUnit.DEGREE, native, value)
        for native, value in ((0, -100), (1000, 0), (2000, 100), (4000, 300))
    ),
    *(
        (VarUnit.PPM, native, value)
        for native, value in ((0, 0), (1, 1), (100, 100), (1000, 1000))
    ),
    # VarUnit.LUX_T,
)


@pytest.mark.parametrize("unit, native, expected", ROUNDTRIP_TEST_VECTORS)
def test_roundtrip(unit: VarUnit, native: int, expected: VarValue) -> None:
    """Test that variable conversion roundtrips."""
    assert (
        expected
        == VarValue.from_var_unit(
            VarValue.to_var_unit(VarValue.from_native(native), unit), unit, True
        ).to_native()
    )


@pytest.mark.parametrize("unit, native, value", CALIBRATION_TEST_VECTORS)
def test_calibration(unit: VarUnit, native: int, value: int | float) -> None:
    """Test proper calibration of variable conversion."""
    assert value == VarValue.to_var_unit(VarValue.from_native(native), unit)



"""LCN devices: Modules and groups."""

from __future__ import annotations

import asyncio
import logging
from collections.abc import Callable, Sequence
from dataclasses import dataclass
from typing import TYPE_CHECKING, Any, cast

from pypck import inputs, lcn_defs
from pypck.helpers import TaskRegistry
from pypck.lcn_addr import LcnAddr
from pypck.pck_commands import PckGenerator
from pypck.status_requester import StatusRequester

if TYPE_CHECKING:
    from pypck.connection import PchkConnectionManager

_LOGGER = logging.getLogger(__name__)


@dataclass
class Serials:
    """Data class for module serials."""

    hardware_serial: int
    manu: int
    software_serial: int
    hardware_type: lcn_defs.HardwareType


class DeviceConnection:
    """Organizes communication with a specific module/group.

    Sends status requests to the connection and handles status responses.
    """

    def __init__(
        self,
        conn: PchkConnectionManager,
        addr: LcnAddr,
        wants_ack: bool = False,
    ) -> None:
        """Construct AbstractConnection instance."""
        self.conn = conn
        self.addr = addr
        self.wants_ack = wants_ack
        self.serials = Serials(-1, -1, -1, lcn_defs.HardwareType.UNKNOWN)
        self._serials_known = asyncio.Event()

        self.input_callbacks: set[Callable[[inputs.Input], None]] = set()

        # List of queued acknowledge codes from the LCN modules.
        self.acknowledges: asyncio.Queue[lcn_defs.AcknowledgeErrorCode] = (
            asyncio.Queue()
        )

        # StatusRequester
        self.status_requester = StatusRequester(self)
        self.request_lock = asyncio.Lock()

        if self.addr.is_group:
            self.wants_ack = False  # groups do not send acks
            self._serials_known.set()
        else:
            self.task_registry.create_task(self._request_device_properties())

    @property
    def task_registry(self) -> TaskRegistry:
        """Get the task registry."""
        return self.conn.task_registry

    @property
    def seg_id(self) -> int:
        """Get the segment id."""
        return self.addr.seg_id

    @property
    def addr_id(self) -> int:
        """Get the module or group id."""
        return self.addr.addr_id

    @property
    def is_group(self) -> int:
        """Return whether this connection refers to a module or group."""
        return self.addr.is_group

    async def send_command(self, wants_ack: bool, pck: str | bytes) -> bool:
        """Send a command to the module represented by this class.

        :param    bool    wants_ack:    Also send a request for acknowledge.
        :param    str     pck:          PCK command (without header).
        """
        if not self.addr.is_group and wants_ack:
            return await self.send_command_with_ack(pck)

        return await self._send_command(wants_ack, pck)

    async def _send_command(self, wants_ack: bool, pck: str | bytes) -> bool:
        """Send a command to the module represented by this class.

        :param    bool    wants_ack:    Also send a request for acknowledge.
        :param    str     pck:          PCK command (without header).
        """
        header = PckGenerator.generate_address_header(
            self.addr, self.conn.local_seg_id, wants_ack
        )
        if isinstance(pck, str):
            return await self.conn.send_command(header + pck)
        return await self.conn.send_command(header.encode() + pck)

    async def serials_known(self) -> None:
        """Wait until the serials of this device are known."""
        await self._serials_known.wait()

    # ##
    # ## Retry logic if an acknowledge is requested
    # ##

    async def send_command_with_ack(self, pck: str | bytes) -> bool:
        """Send a PCK command and ensure receiving of an acknowledgement.

        Resends the PCK command if no acknowledgement has been received
        within timeout.

        :param    str     pck:          PCK command (without header).
        :returns:    True if acknowledge was received, False otherwise
        :rtype:      bool
        """
        count = 0
        while count < self.conn.settings["NUM_TRIES"]:
            await self._send_command(True, pck)
            try:
                code = await asyncio.wait_for(
                    self.acknowledges.get(),
                    timeout=self.conn.settings["DEFAULT_TIMEOUT"],
                )
            except TimeoutError:
                count += 1
                continue
            if code == lcn_defs.AcknowledgeErrorCode.OK:
                return True
            break
        return False

    async def on_ack(
        self, code: lcn_defs.AcknowledgeErrorCode = lcn_defs.AcknowledgeErrorCode.OK
    ) -> None:
        """Is called whenever an acknowledge is received from the LCN module.

        :param     int    code:           The LCN internal code. -1 means
                                          "positive" acknowledge
        """
        await self.acknowledges.put(code)

    # ##
    # ## Methods for sending PCK commands
    # ##

    async def dim_output(self, output_id: int, percent: float, ramp: int) -> bool:
        """Send a dim command for a single output-port.

        :param    int      output_id:    Output id 0..3
        :param    float    percent:      Brightness in percent 0..100
        :param    int      ramp:         Ramp time in milliseconds

        :returns:    True if command was sent successfully, False otherwise
        :rtype:      bool
        """
        return await self.send_command(
            self.wants_ack, PckGenerator.dim_output(output_id, percent, ramp)
        )

    async def dim_all_outputs(self, percent: float, ramp: int) -> bool:
        """Send a dim command for all output-ports.

        :param    float  percent:           Brightness in percent 0..100
        :param    int    ramp:              Ramp time in milliseconds.
        :param    int    software_serial:   The minimum firmware version expected by
                                            any receiving module.

        :returns:    True if command was sent successfully, False otherwise
        :rtype:      bool
        """
        await self._serials_known.wait()
        return await self.send_command(
            self.wants_ack,
            PckGenerator.dim_all_outputs(percent, ramp, self.serials.software_serial),
        )

    async def rel_output(self, output_id: int, percent: float) -> bool:
        """Send a command to change the value of an output-port.

        :param     int    output_id:    Output id 0..3
        :param     float    percent:      Relative brightness in percent
                                        -100..100

        :returns:    True if command was sent successfully, False otherwise
        :rtype:      bool
        """
        return await self.send_command(
            self.wants_ack, PckGenerator.rel_output(output_id, percent)
        )

    async def toggle_output(
        self, output_id: int, ramp: int, to_memory: bool = False
    ) -> bool:
        """Send a command that toggles a single output-port.

        Toggle mode: (on->off, off->on).

        :param    int    output_id:    Output id 0..3
        :param    int    ramp:         Ramp time in milliseconds
        :param    bool   to_memory:    If True, the dimming status is stored

        :returns:    True if command was sent successfully, False otherwise
        :rtype:      bool
        """
        return await self.send_command(
            self.wants_ack, PckGenerator.toggle_output(output_id, ramp, to_memory)
        )

    async def toggle_all_outputs(self, ramp: int) -> bool:
        """Generate a command that toggles all output-ports.

        Toggle Mode:  (on->off, off->on).

        :param    int    ramp:        Ramp time in milliseconds

        :returns:    True if command was sent successfully, False otherwise
        :rtype:      bool
        """
        return await self.send_command(
            self.wants_ack, PckGenerator.toggle_all_outputs(ramp)
        )

    async def control_relays(self, states: list[lcn_defs.RelayStateModifier]) -> bool:
        """Send a command to control relays.

        :param    states:   The 8 modifiers for the relay states as alist
        :type     states:   list(:class:`~pypck.lcn_defs.RelayStateModifier`)

        :returns:    True if command was sent successfully, False otherwise
        :rtype:      bool
        """
        return await self.send_command(
            self.wants_ack, PckGenerator.control_relays(states)
        )

    async def control_relays_timer(
        self, time_msec: int, states: list[lcn_defs.RelayStateModifier]
    ) -> bool:
        """Send a command to control relays.

        :param      int     time_msec:  Duration of timer in milliseconds
        :param    states:   The 8 modifiers for the relay states as alist
        :type     states:   list(:class:`~pypck.lcn_defs.RelayStateModifier`)

        :returns:    True if command was sent successfully, False otherwise
        :rtype:      bool
        """
        return await self.send_command(
            self.wants_ack, PckGenerator.control_relays_timer(time_msec, states)
        )

    async def control_motor_relays(
        self,
        motor_id: int,
        state: lcn_defs.MotorStateModifier,
        mode: lcn_defs.MotorPositioningMode = lcn_defs.MotorPositioningMode.NONE,
    ) -> bool:
        """Send a command to control motors via relays.

        :param    int                    motor_id:    The motor id 0..3
        :param    MotorStateModifier     state:       The modifier for the
        :param    MotorPositioningMode   mode:        The motor positioning mode (ooptional)

        :returns:    True if command was sent successfully, False otherwise
        :rtype:      bool
        """
        return await self.send_command(
            self.wants_ack, PckGenerator.control_motor_relays(motor_id, state, mode)
        )

    async def control_motor_relays_position(
        self,
        motor_id: int,
        position: float,
        mode: lcn_defs.MotorPositioningMode,
    ) -> bool:
        """Control motor position via relays and BS4.

        :param    int                  motor_id:   The motor port of the LCN module
        :param    float                position:   The position to set in percentage (0..100)
        :param    MotorPositioningMode mode:       The motor positioning mode

        :returns:    True if command was sent successfully, False otherwise
        :rtype:      bool
        """
        return await self.send_command(
            self.wants_ack,
            PckGenerator.control_motor_relays_position(motor_id, position, mode),
        )

    async def control_motor_outputs(
        self,
        state: lcn_defs.MotorStateModifier,
        reverse_time: lcn_defs.MotorReverseTime | None = None,
    ) -> bool:
        """Send a command to control a motor via output ports 1+2.

        :param    MotorStateModifier  state: The modifier for the cover state
        :param    MotorReverseTime    reverse_time: Reverse time for modules
                                                    with FW<190C
        :type     state:   :class: `~pypck.lcn-defs.MotorStateModifier`

        :returns:    True if command was sent successfully, False otherwise
        :rtype:      bool
        """
        return await self.send_command(
            self.wants_ack,
            PckGenerator.control_motor_outputs(state, reverse_time),
        )

    async def activate_scene(
        self,
        register_id: int,
        scene_id: int,
        output_ports: Sequence[lcn_defs.OutputPort] = (),
        relay_ports: Sequence[lcn_defs.RelayPort] = (),
        ramp: int | None = None,
    ) -> bool:
        """Activate the stored states for the given scene.

        :param    int                register_id:    Register id 0..9
        :param    int                scene_id:       Scene id 0..9
        :param    list(OutputPort)   output_ports:   Output ports to activate
                                                     as list
        :param    list(RelayPort)    relay_ports:    Relay ports to activate
                                                     as list
        :param    int                ramp:           Ramp value

        :returns:    True if command was sent successfully, False otherwise
        :rtype:      bool
        """
        success = await self.send_command(
            self.wants_ack, PckGenerator.change_scene_register(register_id)
        )
        if not success:
            return False

        result = True
        if output_ports:
            result &= await self.send_command(
                self.wants_ack,
                PckGenerator.activate_scene_output(scene_id, output_ports, ramp),
            )
        if relay_ports:
            result &= await self.send_command(
                self.wants_ack,
                PckGenerator.activate_scene_relay(scene_id, relay_ports),
            )
        return result

    async def store_scene(
        self,
        register_id: int,
        scene_id: int,
        output_ports: Sequence[lcn_defs.OutputPort] = (),
        relay_ports: Sequence[lcn_defs.RelayPort] = (),
        ramp: int | None = None,
    ) -> bool:
        """Store states in the given scene.

        :param    int                register_id:    Register id 0..9
        :param    int                scene_id:       Scene id 0..9
        :param    list(OutputPort)   output_ports:   Output ports to store
                                                     as list
        :param    list(RelayPort)    relay_ports:    Relay ports to store
                                                     as list
        :param    int                ramp:           Ramp value

        :returns:    True if command was sent successfully, False otherwise
        :rtype:      bool
        """
        success = await self.send_command(
            self.wants_ack, PckGenerator.change_scene_register(register_id)
        )

        if not success:
            return False

        result = True
        if output_ports:
            result &= await self.send_command(
                self.wants_ack,
                PckGenerator.store_scene_output(scene_id, output_ports, ramp),
            )
        if relay_ports:
            result &= await self.send_command(
                self.wants_ack,
                PckGenerator.store_scene_relay(scene_id, relay_ports),
            )
        return result

    async def store_scene_outputs_direct(
        self,
        register_id: int,
        scene_id: int,
        percents: Sequence[float],
        ramps: Sequence[int],
    ) -> bool:
        """Store the given output values and ramps in the given scene.

        :param    int           register_id: Register id 0..9
        :param    int           scene_id:    Scene id 0..9
        :param    list(float)   percents:    Output values in percent as list
        :param    list(int)     ramp:        Ramp values as list

        :returns:    True if command was sent successfully, False otherwise
        :rtype:      bool
        """
        return await self.send_command(
            self.wants_ack,
            PckGenerator.store_scene_outputs_direct(
                register_id, scene_id, percents, ramps
            ),
        )

    async def var_abs(
        self,
        var: lcn_defs.Var,
        value: float | lcn_defs.VarValue,
        unit: lcn_defs.VarUnit = lcn_defs.VarUnit.NATIVE,
        software_serial: int = -1,
    ) -> bool:
        """Send a command to set the absolute value to a variable.

        :param     Var        var:      Variable
        :param     float      value:    Absolute value to set
        :param     VarUnit    unit:     Unit of variable

        :returns:    True if command was sent successfully, False otherwise
        :rtype:      bool
        """
        if self.addr.is_group:
            result = True
            # for new modules (>=0x170206)
            result &= await self.var_abs(var, value, unit, 0x170206)

            # for old modules (<0x170206)
            if var in [
                lcn_defs.Var.TVAR,
                lcn_defs.Var.R1VAR,
                lcn_defs.Var.R2VAR,
                lcn_defs.Var.R1VARSETPOINT,
                lcn_defs.Var.R2VARSETPOINT,
            ]:
                result &= await self.var_abs(var, value, unit, 0x000000)
            return result

        if not isinstance(value, lcn_defs.VarValue):
            value = lcn_defs.VarValue.from_var_unit(value, unit, True)

        if software_serial == -1:
            await self._serials_known.wait()
            software_serial = self.serials.software_serial

        if lcn_defs.Var.to_var_id(var) != -1:
            # Absolute commands for variables 1-12 are not supported
            if self.addr_id == 4 and self.is_group:
                # group 4 are status messages
                return await self.send_command(
                    self.wants_ack,
                    PckGenerator.update_status_var(var, value.to_native()),
                )
            # We fake the missing command by using reset and relative
            # commands.
            success = await self.send_command(
                self.wants_ack,
                PckGenerator.var_reset(var, software_serial),
            )
            if not success:
                return False
            return await self.send_command(
                self.wants_ack,
                PckGenerator.var_rel(
                    var,
                    lcn_defs.RelVarRef.CURRENT,
                    value.to_native(),
                    software_serial,
                ),
            )
        return await self.send_command(
            self.wants_ack, PckGenerator.var_abs(var, value.to_native())
        )

    async def var_reset(self, var: lcn_defs.Var, software_serial: int = -1) -> bool:
        """Send a command to reset the variable value.

        :param    Var    var:    Variable

        :returns:    True if command was sent successfully, False otherwise
        :rtype:      bool
        """
        if self.addr.is_group:
            result = True
            result &= await self.var_reset(var, 0x170206)
            if var in [
                lcn_defs.Var.TVAR,
                lcn_defs.Var.R1VAR,
                lcn_defs.Var.R2VAR,
                lcn_defs.Var.R1VARSETPOINT,
                lcn_defs.Var.R2VARSETPOINT,
            ]:
                result &= await self.var_reset(var, 0)
            return result

        if software_serial == -1:
            await self._serials_known.wait()
            software_serial = self.serials.software_serial

        return await self.send_command(
            self.wants_ack, PckGenerator.var_reset(var, software_serial)
        )

    async def var_rel(
        self,
        var: lcn_defs.Var,
        value: float | lcn_defs.VarValue,
        unit: lcn_defs.VarUnit = lcn_defs.VarUnit.NATIVE,
        value_ref: lcn_defs.RelVarRef = lcn_defs.RelVarRef.CURRENT,
        software_serial: int = -1,
    ) -> bool:
        """Send a command to change the value of a variable.

        :param     Var        var:      Variable
        :param     float      value:    Relative value to add (may also be
                                        negative)
        :param     VarUnit    unit:     Unit of variable

        :returns:    True if command was sent successfully, False otherwise
        :rtype:      bool
        """
        if self.addr.is_group:
            result = True
            result &= await self.var_rel(var, value, software_serial=0x170206)
            if var in [
                lcn_defs.Var.TVAR,
                lcn_defs.Var.R1VAR,
                lcn_defs.Var.R2VAR,
                lcn_defs.Var.R1VARSETPOINT,
                lcn_defs.Var.R2VARSETPOINT,
                lcn_defs.Var.THRS1,
                lcn_defs.Var.THRS2,
                lcn_defs.Var.THRS3,
                lcn_defs.Var.THRS4,
                lcn_defs.Var.THRS5,
            ]:
                result &= await self.var_rel(var, value, software_serial=0)
            return result

        if not isinstance(value, lcn_defs.VarValue):
            value = lcn_defs.VarValue.from_var_unit(value, unit, False)

        if software_serial == -1:
            await self._serials_known.wait()
            software_serial = self.serials.software_serial

        return await self.send_command(
            self.wants_ack,
            PckGenerator.var_rel(var, value_ref, value.to_native(), software_serial),
        )

    async def lock_regulator(
        self, reg_id: int, state: bool, target_value: float = -1
    ) -> bool:
        """Send a command to lock a regulator.

        :param    int        reg_id:        Regulator id
        :param    bool       state:         Lock state (locked=True,
                                            unlocked=False)
        :param    float        target_value:  Target value in percent (use -1 to ignore)
        :returns:    True if command was sent successfully, False otherwise
        :rtype:      bool
        """
        return await self.send_command(
            self.wants_ack,
            PckGenerator.lock_regulator(
                reg_id, state, self.serials.software_serial, target_value
            ),
        )

    async def lock_thresholds(
        self, reg_id: int, states: list[lcn_defs.ThresholdLockStateModifier]
    ) -> bool:
        """Send a command to lock thresholds.

        :param    int    reg_id:    Register id 0..3
        :param    list(ThresholdLockStateModifier) states: The 4 threshold lock modifiers as list
        :returns:    True if command was sent successfully, False otherwise
        :rtype:      bool
        """
        return await self.send_command(
            self.wants_ack, PckGenerator.lock_thresholds(reg_id, states)
        )

    async def control_led(
        self, led: lcn_defs.LedPort, state: lcn_defs.LedStatus
    ) -> bool:
        """Send a command to control a led.

        :param    LedPort      led:        Led port
        :param    LedStatus    state:      Led status
        """
        return await self.send_command(
            self.wants_ack, PckGenerator.control_led(led.value, state)
        )

    async def send_keys(
        self, keys: list[list[bool]], cmd: lcn_defs.SendKeyCommand
    ) -> list[bool]:
        """Send a command to send keys.

        :param    list(bool)[4][8]    keys:    2d-list with [table_id][key_id]
                                               bool values, if command should
                                               be sent to specific key
        :param    SendKeyCommand      cmd:     command to send for each table

        :returns:    True if command was sent successfully, False otherwise
        :rtype:      list of bool
        """
        results: list[bool] = []
        for table_id, key_states in enumerate(keys):
            if True in key_states:
                cmds = [lcn_defs.SendKeyCommand.DONTSEND] * 4
                cmds[table_id] = cmd
                results.append(
                    await self.send_command(
                        self.wants_ack, PckGenerator.send_keys(cmds, key_states)
                    )
                )
        return results

    async def send_keys_hit_deferred(
        self, keys: list[list[bool]], delay_time: int, delay_unit: lcn_defs.TimeUnit
    ) -> list[bool]:
        """Send a command to send keys deferred.

        :param    list(bool)[4][8]    keys:          2d-list with
                                                     [table_id][key_id] bool
                                                     values, if command should
                                                     be sent to specific key
        :param    int                 delay_time:    Delay time
        :param    TimeUnit            delay_unit:    Unit of time

        :returns:    True if command was sent successfully, False otherwise
        :rtype:      list of bool
        """
        results: list[bool] = []
        for table_id, key_states in enumerate(keys):
            if True in key_states:
                results.append(
                    await self.send_command(
                        self.wants_ack,
                        PckGenerator.send_keys_hit_deferred(
                            table_id, delay_time, delay_unit, key_states
                        ),
                    ),
                )
        return results

    async def lock_keys(
        self, table_id: int, states: list[lcn_defs.KeyLockStateModifier]
    ) -> bool:
        """Send a command to lock keys.

        :param    int                     table_id:  Table id: 0..3
        :param    keyLockStateModifier    states:    The 8 modifiers for the
                                                     key lock states as a list

        :returns:    True if command was sent successfully, False otherwise
        :rtype:      bool
        """
        return await self.send_command(
            self.wants_ack, PckGenerator.lock_keys(table_id, states)
        )

    async def lock_keys_tab_a_temporary(
        self,
        delay_time: int,
        delay_unit: lcn_defs.TimeUnit,
        states: list[lcn_defs.KeyLockStateModifier],
    ) -> bool:
        """Send a command to lock keys in table A temporary.

        :param    int        delay_time:    Time to lock keys
        :param    TimeUnit   delay_unit:    Unit of time
        :param    list(bool) states:        The 8 lock states of the keys as
                                            list (locked=True, unlocked=False)

        :returns:    True if command was sent successfully, False otherwise
        :rtype:      bool
        """
        return await self.send_command(
            self.wants_ack,
            PckGenerator.lock_keys_tab_a_temporary(delay_time, delay_unit, states),
        )

    async def clear_dyn_text(self, row_id: int) -> bool:
        """Clear previously sent dynamic text.

        :param    int    row_id:    Row id 0..3

        :returns:    True if command was sent successfully, False otherwise
        :rtype:      bool
        """
        return await self.dyn_text(row_id, "")

    async def dyn_text(self, row_id: int, text: str) -> bool:
        """Send dynamic text to a module.

        :param    int    row_id:    Row id 0..3
        :param    str    text:      Text to send (up to 60 bytes)

        :returns:    True if command was sent successfully, False otherwise
        :rtype:      bool
        """
        encoded_text = text.encode(lcn_defs.LCN_ENCODING)
        parts = [encoded_text[12 * part : 12 * part + 12] for part in range(5)]
        result = True
        for part_id, part in enumerate(parts):
            result &= await self.send_command(
                self.wants_ack,
                PckGenerator.dyn_text_part(row_id, part_id, part),
            )
        return result

    async def beep(self, sound: lcn_defs.BeepSound, count: int) -> bool:
        """Send a command to make count number of beep sounds.

        :param    BeepSound sound:  Beep sound style
        :param    int       count:  Number of beeps (1..15)

        :returns:    True if command was sent successfully, False otherwise
        :rtype:      bool
        """
        return await self.send_command(self.wants_ack, PckGenerator.beep(sound, count))

    async def ping(self) -> bool:
        """Send a command that does nothing and request an acknowledgement."""
        return await self.send_command(True, PckGenerator.empty())

    async def pck(self, pck: str) -> bool:
        """Send arbitrary PCK command.

        :param    str    pck:    PCK command

        :returns:    True if command was sent successfully, False otherwise
        :rtype:      bool
        """
        return await self.send_command(self.wants_ack, pck)

    # ##
    # ## Methods for handling input objects
    # ##

    def register_for_inputs(
        self, callback: Callable[[inputs.Input], None]
    ) -> Callable[..., None]:
        """Register a function for callback on PCK message received.

        Returns a function to unregister the callback.
        """
        self.input_callbacks.add(callback)
        return lambda callback=callback: self.input_callbacks.remove(callback)

    async def async_process_input(self, inp: inputs.Input) -> None:
        """Is called by input object's process method.

        Method to handle incoming commands for this specific module (status,
        toggle_output, switch_relays, ...)
        """
        if isinstance(inp, inputs.ModAck):
            await self.on_ack(inp.code)
            return None

        for input_callback in self.input_callbacks:
            input_callback(inp)

    async def dump_details(self) -> dict[str, Any]:
        """Dump detailed information about this module."""
        is_local_segment = self.addr.seg_id in (0, self.conn.local_seg_id)
        return {
            "segment": self.addr.seg_id,
            "address": self.addr.addr_id,
            "is_local_segment": is_local_segment,
            "serials": {
                "hardware_serial": f"{self.serials.hardware_serial:10X}",
                "manu": f"{self.serials.manu:02X}",
                "software_serial": f"{self.serials.software_serial:06X}",
                "hardware_type": f"{self.serials.hardware_type.value:d}",
                "hardware_name": self.serials.hardware_type.description,
            },
            "name": await self.request_name(),
            "comment": await self.request_comment(),
            "oem_text": await self.request_oem_text(),
            "groups": {
                "static": sorted(
                    addr.addr_id
                    for addr in (
                        await self.request_group_memberships(dynamic=False) or set()
                    )
                ),
                "dynamic": sorted(
                    addr.addr_id
                    for addr in (
                        await self.request_group_memberships(dynamic=True) or set()
                    )
                ),
            },
        }

    # ##
    # ## Methods for requesting module properties and status
    # ##

    async def _request_device_properties(self) -> None:
        """Request module properties (serials)."""
        self.serials = await self.request_serials()
        self._serials_known.set()

    # Request status methods

    async def request_status_output(
        self, output_port: lcn_defs.OutputPort, max_age: int = 0
    ) -> inputs.ModStatusOutput | None:
        """Request the status of an output port from a module."""
        if self.addr.is_group:
            _LOGGER.info("Status requests are not supported for groups.")
            return None

        return await self.status_requester.request(
            response_type=inputs.ModStatusOutput,
            request_pck=PckGenerator.request_output_status(output_id=output_port.value),
            max_age=max_age,
            output_id=output_port.value,
        )

    async def request_status_relays(
        self, max_age: int = 0
    ) -> inputs.ModStatusRelays | None:
        """Request the status of relays from a module."""
        if self.addr.is_group:
            _LOGGER.info("Status requests are not supported for groups.")
            return None

        return await self.status_requester.request(
            response_type=inputs.ModStatusRelays,
            request_pck=PckGenerator.request_relays_status(),
            max_age=max_age,
        )

    async def request_status_motor_position(
        self,
        motor: lcn_defs.MotorPort,
        positioning_mode: lcn_defs.MotorPositioningMode,
        max_age: int = 0,
    ) -> inputs.ModStatusMotorPositionBS4 | None:
        """Request the status of motor positions from a module."""
        if self.addr.is_group:
            _LOGGER.info("Status requests are not supported for groups.")
            return None

        if motor not in (
            lcn_defs.MotorPort.MOTOR1,
            lcn_defs.MotorPort.MOTOR2,
            lcn_defs.MotorPort.MOTOR3,
            lcn_defs.MotorPort.MOTOR4,
        ):
            _LOGGER.debug(
                "Only MOTOR1 to MOTOR4 are supported for motor position requests."
            )
            return None
        if positioning_mode != lcn_defs.MotorPositioningMode.BS4:
            _LOGGER.debug("Only BS4 mode is supported for motor position requests.")
            return None

        return await self.status_requester.request(
            response_type=inputs.ModStatusMotorPositionBS4,
            request_pck=PckGenerator.request_motor_position_status(motor.value // 2),
            max_age=max_age,
            motor=motor.value,
        )

    async def request_status_binary_sensors(
        self, max_age: int = 0
    ) -> inputs.ModStatusBinSensors | None:
        """Request the status of binary sensors from a module."""
        if self.addr.is_group:
            _LOGGER.info("Status requests are not supported for groups.")
            return None

        return await self.status_requester.request(
            response_type=inputs.ModStatusBinSensors,
            request_pck=PckGenerator.request_bin_sensors_status(),
            max_age=max_age,
        )

    async def request_status_variable(
        self,
        variable: lcn_defs.Var,
        max_age: int = 0,
    ) -> inputs.ModStatusVar | None:
        """Request the status of a variable from a module."""
        if self.addr.is_group:
            _LOGGER.info("Status requests are not supported for groups.")
            return None

        response_variable = variable

        # for old modules the variable response is typeless
        # - do not use concurrent requests
        # - do not use buffered response
        if has_typeless_response := not lcn_defs.Var.has_type_in_response(
            variable, self.serials.software_serial
        ):
            await self.request_lock.acquire()
            max_age = 0
            response_variable = lcn_defs.Var.UNKNOWN

        result = await self.status_requester.request(
            response_type=inputs.ModStatusVar,
            request_pck=PckGenerator.request_var_status(
                variable, self.serials.software_serial
            ),
            max_age=max_age,
            var=response_variable,
        )

        # for old modules (typeless response) we need to set the original variable
        # - call input_callbacks with the original variable type
        if result is not None and has_typeless_response:
            result.var = variable
            for input_callback in self.input_callbacks:
                input_callback(result)

        if self.request_lock.locked():
            self.request_lock.release()
        return result

    async def request_status_leds_and_logic_ops(
        self, max_age: int = 0
    ) -> inputs.ModStatusLedsAndLogicOps | None:
        """Request the status of LEDs and logic operations from a module."""
        if self.addr.is_group:
            _LOGGER.info("Status requests are not supported for groups.")
            return None

        return await self.status_requester.request(
            response_type=inputs.ModStatusLedsAndLogicOps,
            request_pck=PckGenerator.request_leds_and_logic_ops(),
            max_age=max_age,
        )

    async def request_status_locked_keys(
        self, max_age: int = 0
    ) -> inputs.ModStatusKeyLocks | None:
        """Request the status of locked keys from a module."""
        if self.addr.is_group:
            _LOGGER.info("Status requests are not supported for groups.")
            return None

        return await self.status_requester.request(
            response_type=inputs.ModStatusKeyLocks,
            request_pck=PckGenerator.request_key_lock_status(),
            max_age=max_age,
        )

    # Request module properties

    async def request_serials(self, max_age: int = 0) -> Serials:
        """Request module serials."""
        if self.addr.is_group:
            _LOGGER.info("Status requests are not supported for groups.")
            return Serials(-1, -1, -1, lcn_defs.HardwareType.UNKNOWN)

        result = await self.status_requester.request(
            response_type=inputs.ModSn,
            request_pck=PckGenerator.request_serial(),
            max_age=max_age,
        )

        if isinstance(result, inputs.ModSn):
            return Serials(
                result.hardware_serial,
                result.manu,
                result.software_serial,
                result.hardware_type,
            )
        return Serials(-1, -1, -1, lcn_defs.HardwareType.UNKNOWN)

    async def request_name(self, max_age: int = 0) -> str | None:
        """Request module name."""
        if self.addr.is_group:
            _LOGGER.info("Status requests are not supported for groups.")
            return None

        coros = (
            self.status_requester.request(
                response_type=inputs.ModNameComment,
                request_pck=PckGenerator.request_name(block_id),
                max_age=max_age,
                command="N",
                block_id=block_id,
            )
            for block_id in (0, 1)
        )

        coro_results = [await coro for coro in coros]
        if not all(coro_results):
            return None
        results = cast(list[inputs.ModNameComment], coro_results)
        name = "".join(f"{result.text:10}" for result in results if result)
        return name.rstrip()

    async def request_comment(self, max_age: int = 0) -> str | None:
        """Request module name."""
        if self.addr.is_group:
            _LOGGER.info("Status requests are not supported for groups.")
            return None

        coros = (
            self.status_requester.request(
                response_type=inputs.ModNameComment,
                request_pck=PckGenerator.request_comment(block_id),
                max_age=max_age,
                command="K",
                block_id=block_id,
            )
            for block_id in (0, 1, 2)
        )

        coro_results = [await coro for coro in coros]
        if not all(coro_results):
            return None
        results = cast(list[inputs.ModNameComment], coro_results)
        comment = "".join(f"{result.text:12}" for result in results if result)
        return comment.rstrip()

    async def request_oem_text(self, max_age: int = 0) -> str | None:
        """Request module name."""
        if self.addr.is_group:
            _LOGGER.info("Status requests are not supported for groups.")
            return None

        coros = (
            self.status_requester.request(
                response_type=inputs.ModNameComment,
                request_pck=PckGenerator.request_oem_text(block_id),
                max_age=max_age,
                command="O",
                block_id=block_id,
            )
            for block_id in (0, 1, 2, 3)
        )

        coro_results = [await coro for coro in coros]
        if not all(coro_results):
            return None
        results = cast(list[inputs.ModNameComment], coro_results)
        oem_text = "".join(f"{result.text:12}" for result in results if result)
        return oem_text.rstrip()

    async def request_group_memberships(
        self, dynamic: bool = False, max_age: int = 0
    ) -> set[LcnAddr] | None:
        """Request module static/dynamic group memberships."""
        if self.addr.is_group:
            _LOGGER.info("Status requests are not supported for groups.")
            return set()

        result = await self.status_requester.request(
            response_type=inputs.ModStatusGroups,
            request_pck=(
                PckGenerator.request_group_membership_dynamic()
                if dynamic
                else PckGenerator.request_group_membership_static()
            ),
            max_age=max_age,
            dynamic=dynamic,
        )
        if result is not None:
            return set(result.groups)
        return None



"""Definitions and constants for pypck."""

from __future__ import annotations

import math
import re
from enum import Enum, auto
from typing import Any, no_type_check

LCN_ENCODING = "utf-8"
PATTERN_SPLIT_PORT_PIN = re.compile(r"(?P<port>[a-zA-Z]+)(?P<pin>\d+)")


def split_port_pin(portpin: str) -> tuple[str, int]:
    """Split the port and the pin from the given input string.

    :param    str    portpin:    Input string
    """
    res = PATTERN_SPLIT_PORT_PIN.findall(portpin)
    return res[0][0], int(res[0][1])


class OutputPort(Enum):
    """Output port of LCN module."""

    OUTPUT1 = 0
    OUTPUT2 = 1
    OUTPUT3 = 2
    OUTPUT4 = 3

    OUTPUTUP = 0
    OUTPUTDOWN = 1


class RelayPort(Enum):
    """Relay port of LCN module."""

    RELAY1 = 0
    RELAY2 = 1
    RELAY3 = 2
    RELAY4 = 3
    RELAY5 = 4
    RELAY6 = 5
    RELAY7 = 6
    RELAY8 = 7

    MOTORONOFF1 = 0
    MOTORUPDOWN1 = 1
    MOTORONOFF2 = 2
    MOTORUPDOWN2 = 3
    MOTORONOFF3 = 4
    MOTORUPDOWN3 = 5
    MOTORONOFF4 = 6
    MOTORUPDOWN4 = 7


class MotorPort(Enum):
    """Motor ports of LCN module."""

    MOTOR1 = 0
    MOTOR2 = 1
    MOTOR3 = 2
    MOTOR4 = 3
    OUTPUTS = 4


class LedPort(Enum):
    """LED port of LCN module."""

    LED1 = 0
    LED2 = 1
    LED3 = 2
    LED4 = 3
    LED5 = 4
    LED6 = 5
    LED7 = 6
    LED8 = 7
    LED9 = 8
    LED10 = 9
    LED11 = 10
    LED12 = 11


class LogicOpPort(Enum):
    """Logic Operation port of LCN module."""

    LOGICOP1 = 0
    LOGICOP2 = 1
    LOGICOP3 = 2
    LOGICOP4 = 3


class BinSensorPort(Enum):
    """Binary sensor port of LCN module."""

    BINSENSOR1 = 0
    BINSENSOR2 = 1
    BINSENSOR3 = 2
    BINSENSOR4 = 3
    BINSENSOR5 = 4
    BINSENSOR6 = 5
    BINSENSOR7 = 6
    BINSENSOR8 = 7


class Key(Enum):
    """Keys of LCN module."""

    A1 = 0
    A2 = 1
    A3 = 2
    A4 = 3
    A5 = 4
    A6 = 5
    A7 = 6
    A8 = 7

    B1 = 8
    B2 = 9
    B3 = 10
    B4 = 11
    B5 = 12
    B6 = 13
    B7 = 14
    B8 = 15

    C1 = 16
    C2 = 17
    C3 = 18
    C4 = 19
    C5 = 20
    C6 = 21
    C7 = 22
    C8 = 23

    D1 = 24
    D2 = 25
    D3 = 26
    D4 = 27
    D5 = 28
    D6 = 29
    D7 = 30
    D8 = 31


class KeyAction(Enum):
    """Action types for LCN keys."""

    HIT = "hit"
    MAKE = "make"
    BREAK = "break"


class BatteryStatus(Enum):
    """Battery status."""

    WEAK = "weak"
    FULL = "full"


class OutputPortDimMode(Enum):
    """LCN dimming mode.

    If solely modules with firmware 170206 or newer are present, LCN-PRO
    automatically programs STEPS200.
    Otherwise the default is STEPS50.
    Since LCN-PCHK doesn't know the current mode, it must explicitly be set.
    """

    STEPS50 = 0  # 0..50 dimming steps (all LCN module generations)
    STEPS200 = 1  # 0..200 dimming steps (since 170206)


class OutputPortStatusMode(Enum):
    """Tells LCN-PCHK how to format output-port status-messages.

    PERCENT: allows to show the status in half-percent steps (e.g. "10.5").
    NATIVE: is completely backward compatible and there are no restrictions
    concerning the LCN module generations. It requires LCN-PCHK 2.3 or higher
    though.
    """

    PERCENT = "P"  # Default (compatible with all versions of LCN-PCHK)
    NATIVE = "N"  # 0..200 steps (since LCN-PCHK 2.3)


def time_to_ramp_value(time_msec: int) -> int:
    """Convert the given time into an LCN ramp value.

    :param    int    time_msec:    The time in milliseconds.

    :returns: The (LCN-internal) ramp value (0..250).
    :rtype: int
    """
    if time_msec < 250:
        ret = 0
    elif time_msec < 500:
        ret = 1
    elif time_msec < 660:
        ret = 2
    elif time_msec < 1000:
        ret = 3
    elif time_msec < 1400:
        ret = 4
    elif time_msec < 2000:
        ret = 5
    elif time_msec < 3000:
        ret = 6
    elif time_msec < 4000:
        ret = 7
    elif time_msec < 5000:
        ret = 8
    elif time_msec < 6000:
        ret = 9
    else:
        ramp = (time_msec / 1000 - 6) / 2 + 10
        ramp = min(ramp, 250)
        ret = int(ramp)
    return ret


def ramp_value_to_time(ramp_value: int) -> int:
    """Convert the given LCN ramp value into a time.

    :param    int    ramp_value:    The LCN ramp value (0..250).

    :returns: The ramp time in milliseconds.
    :rtype: int
    """
    if not 0 <= ramp_value <= 250:
        raise ValueError("Ramp value has to be in range 0..250")

    if ramp_value < 10:
        times = [0, 250, 500, 660, 1000, 1400, 2000, 3000, 4000, 5000]
        ramp_time = times[ramp_value]
    else:
        ramp_time = int(((ramp_value - 10) * 2 + 6) * 1000)
    return ramp_time


def time_to_native_value(time_msec: int) -> int:
    """Convert time to native LCN time value.

    Scales the given time value in milliseconds to a byte value (0..255).
    Used for RelayTimer.

    :param    int    time_msec:    Duration of timer in milliseconds

    :returns: The duration in native LCN units
    :rtype: int
    """
    if not 0 <= time_msec <= 240960:
        raise ValueError("Time has to be in range 0..240960ms")
    time_scaled = time_msec / (1000 * 0.03 * 32.0) + 1.0

    pre_decimal = int(time_scaled).bit_length() - 1
    decimal = time_scaled / (1 << pre_decimal) - 1

    value = pre_decimal + decimal
    return int(32 * value)


def native_value_to_time(value: int) -> int:
    """Convert native LCN value to time.

    Scales the given byte value (0..255) to a time value in milliseconds.

    :param    int    value:    Duration of timer in native LCN units

    :returns: The duration in milliseconds
    :rtype: int
    """
    if not 0 <= value <= 255:
        raise ValueError("Value has to be in range 0..255")
    pre_decimal = value // 32
    decimal = value / 32 - pre_decimal

    time_scaled = (1 << pre_decimal) * (decimal + 1)

    time_msec = (time_scaled - 1) * 1000 * 0.03 * 32
    return int(time_msec)


def motor_position_time_to_native_value(time_msec: int) -> int:
    """Convert time to native LCN time value.

    Scales the given time value in milliseconds to a two-byte value.

    :param    int    time_msec:    Duration of timer in milliseconds (1001..65535000)

    :returns: The duration in native LCN units
    :rtype: int
    """
    if not 1001 <= time_msec <= 65535000:
        raise ValueError("Time has to be in range 1001..65535000ms")
    value = 0xFFFF * 1000 / time_msec
    return int(value)


def native_value_to_motor_position_time(value: int) -> int:
    """Convert native LCN value to time.

    Scales the given two-byte value (1..65535) to a time value in milliseconds.

    :param    int    value:    Duration of timer in native LCN units

    :returns: The duration in milliseconds
    :rtype: int
    """
    if not 1 <= value <= 0xFFFF:
        raise ValueError("Value has to be in range 1..65535")

    time_msec = 0xFFFF * 1000 / value
    return int(time_msec)


class Var(Enum):
    """LCN variable types."""

    UNKNOWN = -1  # Used if the real type is not known (yet)
    VAR1ORTVAR = 0
    TVAR = 0
    VAR1 = 0
    VAR2ORR1VAR = 1
    R1VAR = 1
    VAR2 = 1
    VAR3ORR2VAR = 2
    R2VAR = 2
    VAR3 = 2
    VAR4 = 3
    VAR5 = 4
    VAR6 = 5
    VAR7 = 6
    VAR8 = 7
    VAR9 = 8
    VAR10 = 9
    VAR11 = 10
    VAR12 = 11  # Since 170206
    R1VARSETPOINT = auto()
    R2VARSETPOINT = auto()  # Set-points for regulators
    THRS1 = auto()
    THRS2 = auto()
    THRS3 = auto()
    THRS4 = auto()
    THRS5 = auto()  # Register 1 (THRS5 only before 170206)
    THRS2_1 = auto()
    THRS2_2 = auto()
    THRS2_3 = auto()
    THRS2_4 = auto()  # Register 2 (since 2012)
    THRS3_1 = auto()
    THRS3_2 = auto()
    THRS3_3 = auto()
    THRS3_4 = auto()  # Register 3 (since 2012)
    THRS4_1 = auto()
    THRS4_2 = auto()
    THRS4_3 = auto()
    THRS4_4 = auto()  # Register 4 (since 2012)
    S0INPUT1 = auto()
    S0INPUT2 = auto()
    S0INPUT3 = auto()
    S0INPUT4 = auto()  # LCN-BU4LJVarValue

    @classmethod
    def variables(cls) -> list[Var]:
        """Return a list of all variable types."""
        return [
            cls.VAR1ORTVAR,
            cls.VAR2ORR1VAR,
            cls.VAR3ORR2VAR,
            cls.VAR4,
            cls.VAR5,
            cls.VAR6,
            cls.VAR7,
            cls.VAR8,
            cls.VAR9,
            cls.VAR10,
            cls.VAR11,
            cls.VAR12,
        ]

    @classmethod
    def variables_new(cls) -> list[Var]:
        """Return a list of all new variable types (firmware >=0x170206)."""
        return cls.variables()

    @classmethod
    def variables_old(cls) -> list[Var]:
        """Return a list of all variable types (firmware <0x170206)."""
        return cls.variables()[:3]

    @classmethod
    def set_points(cls) -> list[Var]:
        """Return a list of all set-point variable types."""
        return [cls.R1VARSETPOINT, cls.R2VARSETPOINT]

    @classmethod
    def thresholds(cls) -> list[list[Var]]:
        """Return a list of all threshold variable types."""
        return [
            [cls.THRS1, cls.THRS2, cls.THRS3, cls.THRS4, cls.THRS5],
            [cls.THRS2_1, cls.THRS2_2, cls.THRS2_3, cls.THRS2_4],
            [cls.THRS3_1, cls.THRS3_2, cls.THRS3_3, cls.THRS3_4],
            [cls.THRS4_1, cls.THRS4_2, cls.THRS4_3, cls.THRS4_4],
        ]

    @classmethod
    def thresholds_new(cls) -> list[list[Var]]:
        """Return a list of all threshold variable types (firmware >=0x170206)."""
        return [cls.thresholds()[0][:4], *cls.thresholds()[1:]]

    @classmethod
    def thresholds_old(cls) -> list[list[Var]]:
        """Return a list of all old threshold variable types (firmware <0x170206)."""
        return [cls.thresholds()[0]]

    @classmethod
    def s0s(cls) -> list[Var]:
        """Return a list of all S0-input variable types."""
        return [cls.S0INPUT1, cls.S0INPUT2, cls.S0INPUT3, cls.S0INPUT4]

    @staticmethod
    def var_id_to_var(var_id: int) -> Var:
        """Translate a given id into a variable type.

        :param    int    varId:    The variable id (0..11)

        :returns: The translated variable enum.
        :rtype: Var
        """
        if (var_id < 0) or (var_id >= len(Var.variables())):
            raise ValueError("Bad var_id.")
        return Var.variables()[var_id]

    @staticmethod
    def set_point_id_to_var(set_point_id: int) -> Var:
        """Translate a given id into a LCN set-point variable type.

        :param     int    set_point_id:    Set-point id 0..1

        :return: The translated var
        :rtype:  Var
        """
        if (set_point_id < 0) or (set_point_id >= len(Var.set_points())):
            raise ValueError("Bad set_point_id.")
        return Var.set_points()[set_point_id]

    @staticmethod
    def thrs_id_to_var(register_id: int, thrs_id: int) -> Var:
        """Translate given ids into a LCN threshold variable type.

        :param    int    register_id:    Register id 0..3
        :param    int    thrs_id:        Threshold id 0..4 for register 0,
                                                      0..3 for registers 1..3

        :return: The translated var
        :rtype:    Var
        """
        if (
            (register_id < 0)
            or (register_id >= len(Var.thresholds()))
            or (thrs_id < 0)
            or (thrs_id >= (5 if (register_id == 0) else 4))
        ):
            raise ValueError("Bad register_id and/or thrs_id.")
        return Var.thresholds()[register_id][thrs_id]

    @staticmethod
    def s0_id_to_var(s0_id: int) -> Var:
        """Translate a given id into a LCN S0-input variable type.

        :param     int    s0_id:     S0 id 0..3

        :return:    The translated var
        :rtype:     Var
        """
        if (s0_id < 0) or (s0_id >= len(Var.s0s())):
            raise ValueError("Bad s0_id.")
        return Var.s0s()[s0_id]

    @staticmethod
    def to_var_id(var: Var) -> int:
        """Translate a given variable type into a variable id.

        :param     Var    var:    The variable type to translate

        :return:     Variable id 0..11 or -1 if wrong type
        :rtype:    int
        """
        if var == Var.VAR1ORTVAR:
            var_id = 0
        elif var == Var.VAR2ORR1VAR:
            var_id = 1
        elif var == Var.VAR3ORR2VAR:
            var_id = 2
        elif var == Var.VAR4:
            var_id = 3
        elif var == Var.VAR5:
            var_id = 4
        elif var == Var.VAR6:
            var_id = 5
        elif var == Var.VAR7:
            var_id = 6
        elif var == Var.VAR8:
            var_id = 7
        elif var == Var.VAR9:
            var_id = 8
        elif var == Var.VAR10:
            var_id = 9
        elif var == Var.VAR11:
            var_id = 10
        elif var == Var.VAR12:
            var_id = 11
        else:
            var_id = -1
        return var_id

    @staticmethod
    def to_set_point_id(var: Var) -> int:
        """Translate a given variable type into a set-point id.

        :param     Var    var:     The variable type to translate

        :return:    Variable id 0..1 or -1 if wrong type
        :rtype:    int
        """
        if var == Var.R1VARSETPOINT:
            set_point_id = 0
        elif var == Var.R2VARSETPOINT:
            set_point_id = 1
        else:
            set_point_id = -1
        return set_point_id

    @staticmethod
    def to_thrs_register_id(var: Var) -> int:
        """Translate a given variable type into a threshold register id.

        :param    Var    var:    The variable type to translate

        :return:    Register id 0..3 or -1 if wrong type
        :rtype:    int
        """
        if var in [Var.THRS1, Var.THRS2, Var.THRS3, Var.THRS4, Var.THRS5]:
            thrs_register_id = 0
        elif var in [Var.THRS2_1, Var.THRS2_2, Var.THRS2_3, Var.THRS2_4]:
            thrs_register_id = 1
        elif var in [Var.THRS3_1, Var.THRS3_2, Var.THRS3_3, Var.THRS3_4]:
            thrs_register_id = 2
        elif var in [Var.THRS4_1, Var.THRS4_2, Var.THRS4_3, Var.THRS4_4]:
            thrs_register_id = 3
        else:
            thrs_register_id = -1
        return thrs_register_id

    @staticmethod
    def to_thrs_id(var: Var) -> int:
        """Translate a given variable type into a threshold id.

        :param    Var    var:    The variable type to translate

        :return:    Threshold id 0..4 or -1 if wrong type
        :rtype:    int
        """
        if var in [Var.THRS1, Var.THRS2_1, Var.THRS3_1, Var.THRS4_1]:
            thrs_id = 0
        elif var in [Var.THRS2, Var.THRS2_2, Var.THRS3_2, Var.THRS4_2]:
            thrs_id = 1
        elif var in [Var.THRS3, Var.THRS2_3, Var.THRS3_3, Var.THRS4_3]:
            thrs_id = 2
        elif var in [Var.THRS4, Var.THRS2_4, Var.THRS3_4, Var.THRS4_4]:
            thrs_id = 3
        elif var == Var.THRS5:
            thrs_id = 4
        else:
            thrs_id = -1
        return thrs_id

    @staticmethod
    def to_s0_id(var: Var) -> int:
        """Translate a given variable type into an S0-input id.

        :param    Var    var:    The variable type to translate

        :return:    S0 id 0..3 or -1 if wrong type
        :rtype:    int
        """
        if var == Var.S0INPUT1:
            s0_id = 0
        elif var == Var.S0INPUT2:
            s0_id = 1
        elif var == Var.S0INPUT3:
            s0_id = 2
        elif var == Var.S0INPUT4:
            s0_id = 3
        else:
            s0_id = -1
        return s0_id

    @staticmethod
    def is_lockable_regulator_source(var: Var) -> bool:
        """Check if the the given variable type is lockable.

        :param    Var    var:    The variable type to check

        :return:    True if lockable, otherwise False
        :rtype:    bool
        """
        return var in [Var.R1VARSETPOINT, Var.R2VARSETPOINT]

    @staticmethod
    def use_lcn_special_values(var: Var) -> bool:
        """Check if the given variable type uses special values.

        Examples for special values: 'No value yet', 'sensor defective' etc.

        :param    Var    var:    The variable type to check

        :return:    True if special values are in use, otherwise False
        :rtype:    bool
        """
        return var not in [Var.S0INPUT1, Var.S0INPUT2, Var.S0INPUT3, Var.S0INPUT4]

    @staticmethod
    def has_type_in_response(var: Var, software_serial: int) -> bool:
        """Module-generation check.

        Check if the given variable type would receive a typed response if
        its status was requested.

        :param    Var    var:    The variable type to check
        :param    int    swAge:  The target LCN-modules firmware version

        :return:    True if a response would contain the variable's type,
                    otherwise False
        :rtype:    bool
        """
        if software_serial < 0x170206:
            if var in [
                Var.VAR1ORTVAR,
                Var.VAR2ORR1VAR,
                Var.VAR3ORR2VAR,
                Var.R1VARSETPOINT,
                Var.R2VARSETPOINT,
            ]:
                return False
        return True

    @staticmethod
    def is_event_based(var: Var, software_serial: int) -> bool:
        """Module-generation check.

        Check if the given variable type automatically sends status-updates
        on value-change. It must be polled otherwise.

        :param    Var    var:    The variable type to check
        :param    int    swAge:  The target LCN-module's firmware version

        :return:    True if the LCN module supports automatic status-messages
                    for this var, otherwise False
        :rtype:    bool
        """
        if (Var.to_set_point_id(var) != -1) or (Var.to_s0_id(var) != -1):
            return True
        return software_serial >= 0x170206

    @staticmethod
    def should_poll_status_after_command(var: Var, is2013: bool) -> bool:
        """Module-generation check.

        Check if the target LCN module would automatically send status-updates
        if the given variable type was changed by command.

        :param    Var    var:    The variable type to check
        :param    bool   is2013: The target module's-generation

        :return:    True if a poll is required to get the new status-value,
                    otherwise False
        :rtype:    bool
        """
        # Regulator set-points will send status-messages on every change
        # (all firmware versions)
        if Var.to_set_point_id(var) != -1:
            return False

        # Thresholds since 170206 will send status-messages on every change
        if is2013 and (Var.to_thrs_register_id(var) != -1):
            return False

        # Others:
        # - Variables before 170206 will never send any status-messages
        # - Variables since 170206 only send status-messages on "big" changes
        # - Thresholds before 170206 will never send any status-messages
        # - S0-inputs only send status-messages on "big" changes
        # (all "big changes" cases force us to poll the status to get faster
        # updates)
        return True

    @staticmethod
    def should_poll_status_after_regulator_lock(
        software_serial: int, lock_state: int
    ) -> bool:
        """Module-generation check.

        Check if the target LCN module would automatically send status-updates
        if the given regulator's lock-state was changed by command.

        :param    int       swAge: The target LCN-module's firmware version
        :param    int   lockState: The lock-state sent via command

        :return:    True if a poll is required to get the new status-value,
                    otherwise False
        :rtype:    bool
        """
        # LCN modules before 170206 will send an automatic status-message for
        # "lock", but not for "unlock"
        return (not lock_state) and (software_serial < 0x170206)


class VarUnit(Enum):
    """Measurement units used with LCN variables."""

    NATIVE = ""  # LCN internal representation (0 = -100C for absolute values)
    CELSIUS = "\u00b0C"
    KELVIN = "\u00b0K"
    FAHRENHEIT = "\u00b0F"
    LUX_T = "Lux_T"
    LUX_I = "Lux_I"
    METERPERSECOND = "m/s"  # Used for LCN-WIH wind speed
    PERCENT = "%"  # Used for humidity
    PPM = "ppm"  # Used by CO2 sensor
    VOLT = "V"
    AMPERE = "A"
    DEGREE = "\u00b0"  # Used for angles,

    @staticmethod
    def parse(unit: str) -> VarUnit:
        """Parse the given unit string and return VarUnit.

        :param    str    unit:    The input unit
        """
        unit = unit.upper()
        if unit in ["", "NATIVE", "LCN"]:
            var_unit = VarUnit.NATIVE
        elif unit in ["CELSIUS", "\u00b0CELSIUS", "\u00b0C"]:
            var_unit = VarUnit.CELSIUS
        elif unit in ["KELVIN", "\u00b0KELVIN", "\u00b0K", "K"]:
            var_unit = VarUnit.KELVIN
        elif unit in ["FAHRENHEIT", "\u00b0FAHRENHEIT", "\u00b0F"]:
            var_unit = VarUnit.FAHRENHEIT
        elif unit in ["LUX_T", "LX_T"]:
            var_unit = VarUnit.LUX_T
        elif unit in ["LUX", "LUX_I", "LX"]:
            var_unit = VarUnit.LUX_I
        elif unit in ["M/S", "METERPERSECOND"]:
            var_unit = VarUnit.METERPERSECOND
        elif unit in ["%", "PERCENT"]:
            var_unit = VarUnit.PERCENT
        elif unit == "PPM":
            var_unit = VarUnit.PPM
        elif unit in ["VOLT", "V"]:
            var_unit = VarUnit.VOLT
        elif unit in ["AMPERE", "AMP", "A"]:
            var_unit = VarUnit.AMPERE
        elif unit in ["DEGREE", "\u00b0"]:
            var_unit = VarUnit.DEGREE
        else:
            raise ValueError("Bad input unit.")
        return var_unit


class VarValue:
    """A value of an LCN variable.

    It internally stores the native LCN value and allows to convert from/into
    other units. Some conversions allow to specify whether the source value is
    absolute or relative. Relative values are used to create varvalues that
    can be added/subtracted from other (absolute) varvalues.

    :param    int    native_value:    The native value
    """

    def __init__(self, native_value: int) -> None:
        """Construct with native LCN value."""
        self.native_value = native_value

    def __eq__(self, other: object) -> bool:
        """Return if instance equals the given object."""
        if isinstance(other, VarValue):
            return self.native_value == other.native_value
        return False

    def __hash__(self) -> int:
        """Calculate the instance hash value."""
        return self.native_value.__hash__()

    def is_locked_regulator(self) -> bool:
        """Return if regulator is locked."""
        return self.is_locked()

    def is_locked_threshold(self) -> bool:
        """Return if threshold is locked."""
        return self.is_locked()

    def is_locked(self) -> bool:
        """Return if variable is locked."""
        return (self.native_value & 0x8000) != 0

    @staticmethod
    def from_var_unit(value: float, unit: VarUnit, is_abs: bool) -> VarValue:
        """Create a variable value from any input.

        :param    float    value:   The input value
        :param    VarUnit  unit:    The input value's unit
        :param    bool     is_abs:  True for absolute values (relative values
                                    are used to add/subtract from other
                                    VarValues), otherwise False

        :return:    The variable value (never null)
        :rtype:    VarValue
        """
        if unit == VarUnit.NATIVE:
            var_value = VarValue.from_native(int(value))
        elif unit == VarUnit.CELSIUS:
            var_value = VarValue.from_celsius(value, is_abs)
        elif unit == VarUnit.KELVIN:
            var_value = VarValue.from_kelvin(value, is_abs)
        elif unit == VarUnit.FAHRENHEIT:
            var_value = VarValue.from_fahrenheit(value, is_abs)
        elif unit == VarUnit.LUX_T:
            var_value = VarValue.from_lux_t(value)
        elif unit == VarUnit.LUX_I:
            var_value = VarValue.from_lux_i(value)
        elif unit == VarUnit.METERPERSECOND:
            var_value = VarValue.from_meters_per_second(value)
        elif unit == VarUnit.PERCENT:
            var_value = VarValue.from_percent(value)
        elif unit == VarUnit.PPM:
            var_value = VarValue.from_ppm(value)
        elif unit == VarUnit.VOLT:
            var_value = VarValue.from_volt(value)
        elif unit == VarUnit.AMPERE:
            var_value = VarValue.from_ampere(value)
        elif unit == VarUnit.DEGREE:
            var_value = VarValue.from_degree(value, is_abs)
        else:
            raise ValueError("Wrong unit.")
        return var_value

    @staticmethod
    def from_native(value: int) -> VarValue:
        """Create a variable value from native input.

        :param    int    value:    The input value

        :return:    The variable value (never null)
        :rtype:    VarValue
        """
        return VarValue(value)

    @staticmethod
    def from_celsius(value: float, is_abs: bool = True) -> VarValue:
        """Create a variable value from Celsius input.

        :param    float    value:    The input value
        :param    bool     is_abs:   True for absolute values (relative values
                                     are used to add/subtract from other
                                     VarValues), otherwise False

        :return:    The variable value (never null)
        :rtype:    VarValue
        """
        number = int(round(value * 10))
        return VarValue(number + 1000 if is_abs else number)

    @staticmethod
    def from_kelvin(value: float, is_abs: bool = True) -> VarValue:
        """Create a variable value from Kelvin input.

        :param    float    value:    The input value
        :param    bool     is_abs:   True for absolute values (relative values
                                     are used to add/subtract from other
                                     VarValues), otherwise False

        :return:    The variable value (never null)
        :rtype:    VarValue
        """
        if is_abs:
            value -= 273.15

        number = int(round(value * 10))
        return VarValue(number + 1000 if is_abs else number)

    @staticmethod
    def from_fahrenheit(value: float, is_abs: bool = True) -> VarValue:
        """Create a variable value from Fahrenheit input.

        :param    float    value:    The input value
        :param    bool     is_abs:   True for absolute values (relative values
                                     are used to add/subtract from other
                                     VarValues), otherwise False

        :return:    The variable value (never null)
        :rtype:    VarValue
        """
        if is_abs:
            value -= 32

        number = int(round(value / 0.18))
        return VarValue(number + 1000 if is_abs else number)

    @staticmethod
    def from_lux_t(lux: float) -> VarValue:
        """Create a variable value from lx input.

        Target must be connected to T-port.

        :param    float    l:    The input value

        :return:    The variable value (never null)
        :rtype:    VarValue
        """
        return VarValue(int(round(math.log(lux) - 1.689646994) / 0.010380664))

    @staticmethod
    def from_lux_i(lux: float) -> VarValue:
        """Create a variable value from lx input.

        Target must be connected to I-port.

        :param    float    l:    The input value

        :return:    The variable value (never null)
        :rtype:    VarValue
        """
        return VarValue(int(round(math.log(lux) * 100)))

    @staticmethod
    def from_percent(value: float) -> VarValue:
        """Create a variable value from % input.

        :param    float    value:    The input value

        :return:    The variable value (never null)
        :rtype:    VarValue
        """
        return VarValue(int(round(value)))

    @staticmethod
    def from_ppm(value: float) -> VarValue:
        """Create a variable value from ppm input.

        Used for CO2 sensors.

        :param    float    value:   The input value

        :return:    The variable value (never null)
        :rtype:    VarValue
        """
        return VarValue(int(round(value)))

    @staticmethod
    def from_meters_per_second(value: float) -> VarValue:
        """Create a variable value from m/s input.

        Used for LCN-WIH wind speed.

        :param    float    value:    The input value

        :return:    The variable value (never null)
        :rtype:    VarValue
        """
        return VarValue(int(round(value * 10)))

    @staticmethod
    def from_volt(value: float) -> VarValue:
        """Create a variable value from V input.

        :param    float    value:    The input value

        :return:    The variable value (never null)
        :rtype:    VarValue
        """
        return VarValue(int(round(value * 400)))

    @staticmethod
    def from_ampere(value: float) -> VarValue:
        """Create a variable value from A input.

        :param    float    value:    The input value

        :return: The variable value (never null)
        :rtype:    VarValue
        """
        return VarValue(int(round(value * 100000)))

    @staticmethod
    def from_degree(value: float, is_abs: bool = True) -> VarValue:
        """Create a variable value from degree (angle) input.

        :param    float    value:        The input value
        :param    bool     is_abs:   True for absolute values (relative values
                                     are used to add/subtract from other
                                     VarValues), otherwise False

        :return:    The variable value (never null)
        :rtype:     VarValue
        """
        number = int(round(value * 10))
        return VarValue(number + 1000 if is_abs else number)

    def to_var_unit(
        self,
        unit: VarUnit,
        is_lockable_regulator_source: bool = False,
    ) -> int | float:
        """Convert the given unit to a VarValue.

        :param    VarUnit    unit:    The variable unit
        :param    bool       is_lockable_regulator_source:  Is lockable source

        :return:    The variable value
        :rtype:     Union[int,float]
        """
        var_value = VarValue(
            self.native_value & 0x7FFF
            if is_lockable_regulator_source
            else self.native_value
        )

        if unit == VarUnit.NATIVE:
            return var_value.to_native()
        if unit == VarUnit.CELSIUS:
            return var_value.to_celsius()
        if unit == VarUnit.KELVIN:
            return var_value.to_kelvin()
        if unit == VarUnit.FAHRENHEIT:
            return var_value.to_fahrenheit()
        if unit == VarUnit.LUX_T:
            return var_value.to_lux_t()
        if unit == VarUnit.LUX_I:
            return var_value.to_lux_i()
        if unit == VarUnit.METERPERSECOND:
            return var_value.to_meters_per_second()
        if unit == VarUnit.PERCENT:
            return var_value.to_percent()
        if unit == VarUnit.PPM:
            return var_value.to_ppm()
        if unit == VarUnit.VOLT:
            return var_value.to_volt()
        if unit == VarUnit.AMPERE:
            return var_value.to_ampere()
        if unit == VarUnit.DEGREE:
            return var_value.to_degree()
        raise ValueError("Wrong unit.")

    def to_native(self) -> int:
        """Convert to native value.

        :return:    The converted value
        :rtype:    int
        """
        return self.native_value

    def to_celsius(self) -> float:
        """Convert to Celsius value.

        :return:    The converted value
        :rtype:    float
        """
        return (self.native_value - 1000) / 10.0

    def to_kelvin(self) -> float:
        """Convert to Kelvin value.

        :return:    The converted value
        :rtype:    float
        """
        return (self.native_value - 1000) / 10.0 + 273.15

    def to_fahrenheit(self) -> float:
        """Convert to Fahrenheit value.

        :return:    The converted value
        :rtype:    float
        """
        return (self.native_value - 1000) * 0.18 + 32.0

    def to_lux_t(self) -> float:
        """Convert to lx value.

        Source must be connected to T-port.

        :return:    The converted value
        :rtype:    float
        """
        return math.exp(0.010380664 * self.native_value + 1.689646994)

    def to_lux_i(self) -> float:
        """Convert to lx value.

        Source must be connected to I-port.

        :return:    The converted value
        :rtype:    float
        """
        return math.exp(self.native_value / 100)

    def to_percent(self) -> int:
        """Convert to % value.

        :return:    The converted value
        :rtype:    int
        """
        return self.native_value

    def to_ppm(self) -> int:
        """Convert to ppm value.

        :return:    The converted value
        :rtype:    int
        """
        return self.native_value

    def to_meters_per_second(self) -> float:
        """Convert to m/s value.

        :return:    The converted value
        :rtype:    float
        """
        return self.native_value / 10.0

    def to_volt(self) -> float:
        """Convert to V value.

        :return:    The converted value
        :rtype:    float
        """
        return self.native_value / 400.0

    def to_ampere(self) -> float:
        """Convert to A value.

        :return:    The converted value
        :rtype:    float
        """
        return self.native_value / 100000.0

    def to_degree(self) -> float:
        """Convert to degree value.

        :return:    The converted value
        :rtype:    float
        """
        return (self.native_value - 1000) / 10.0

    def to_var_unit_string(
        self,
        unit: VarUnit,
        is_lockable_regulator_source: bool = False,
        use_lcn_special_values: bool = False,
    ) -> str:
        """Convert the given unit into a string representation.

        :param    VarUnit    unit:    The input unit
        :param    bool       is_lockable_regulator_source:  Is lockable source
        :param    bool       use_lcn_special_values:  Use LCN special values

        :return:    The string representation of input unit.
        :rtype:     str
        """
        if use_lcn_special_values and (self.native_value == 0xFFFF):  # No value
            ret = "---"
        elif use_lcn_special_values and (
            (self.native_value & 0xFF00) == 0x8100
        ):  # Undefined
            ret = "---"
        elif use_lcn_special_values and (
            (self.native_value & 0xFF00) == 0x7F00
        ):  # Defective
            ret = "!!!"
        else:
            var = VarValue(
                (self.native_value & 0x7FF)
                if is_lockable_regulator_source
                else self.native_value
            )
            if unit == VarUnit.NATIVE:
                ret = f"{var.to_native():.0f}"
            elif unit == VarUnit.CELSIUS:
                ret = f"{var.to_celsius():.01f}"
            elif unit == VarUnit.KELVIN:
                ret = f"{var.to_kelvin():.01f}"
            elif unit == VarUnit.FAHRENHEIT:
                ret = f"{var.to_fahrenheit():.01f}"
            elif unit == VarUnit.LUX_T:
                if var.to_native() > 1152:  # Max. value the HW can do
                    ret = "---"
                else:
                    ret = f"{var.to_lux_t():.0f}"
            elif unit == VarUnit.LUX_I:
                if var.to_native() > 1152:  # Max. value the HW can do
                    ret = "---"
                else:
                    ret = f"{var.to_lux_i():.0f}"
            elif unit == VarUnit.METERPERSECOND:
                ret = f"{var.to_meters_per_second():.0f}"
            elif unit == VarUnit.PERCENT:
                ret = f"{var.to_percent():.0f}"
            elif unit == VarUnit.PPM:
                ret = f"{var.to_ppm():.0f}"
            elif unit == VarUnit.VOLT:
                ret = f"{var.to_volt():.0f}"
            elif unit == VarUnit.AMPERE:
                ret = f"{var.to_ampere():.0f}"
            elif unit == VarUnit.DEGREE:
                ret = f"{var.to_degree():.0f}"
            else:
                raise ValueError("Wrong unit.")

        # handle locked regulators
        if is_lockable_regulator_source and self.is_locked():
            ret = f"({ret:s})"

        return ret


class LedStatus(Enum):
    """Possible states for LCN LEDs."""

    OFF = "A"
    ON = "E"
    BLINK = "B"
    FLICKER = "F"


class LogicOpStatus(Enum):
    """Possible states for LCN logic-operations."""

    NONE = "N"
    SOME = "T"  # Note: Actually not correct since AND won't be OR also
    ALL = "V"


class TimeUnit(Enum):
    """Time units used for several LCN commands."""

    SECONDS = "S"
    MINUTES = "M"
    HOURS = "H"
    DAYS = "D"

    @staticmethod
    def parse(unit: str) -> TimeUnit:
        """Parse the given time_unit into a time unit.

        It supports several alternative terms.

        :param    str    time_unit:    The text to parse

        :return:    TimeUnit enum
        :rtype:    TimeUnit
        """
        unit = unit.upper()
        if unit in ["SECONDS", "SECOND", "SEC", "S"]:
            time_unit = TimeUnit.SECONDS
        elif unit in ["MINUTES", "MINUTE", "MIN", "M"]:
            time_unit = TimeUnit.MINUTES
        elif unit in ["HOURS", "HOUR", "H"]:
            time_unit = TimeUnit.HOURS
        elif unit in ["DAYS", "DAY", "D"]:
            time_unit = TimeUnit.DAYS
        else:
            raise ValueError("Bad time unit input.")
        return time_unit


class RelayStateModifier(Enum):
    """Relay-state modifiers used in LCN commands."""

    ON = "1"
    OFF = "0"
    TOGGLE = "U"
    NOCHANGE = "-"


class MotorStateModifier(Enum):
    """Motor-state modifiers used in LCN commands.

    LCN module has to be configured for motors connected to relays.
    """

    UP = "U"
    DOWN = "D"
    STOP = "S"
    TOGGLEONOFF = "T"  # toggle on/off
    TOGGLEDIR = "R"  # toggle direction
    CYCLE = "C"  # up, stop, down, stop, ...
    NOCHANGE = "-"


class MotorReverseTime(Enum):
    """Motor reverse time user in LCN commands.

    For modules with FW<190C the release time has to be specified.
    """

    RT70 = "RT70"  # 70ms
    RT600 = "RT600"  # 600ms
    RT1200 = "RT1200"  # 1200ms


class MotorPositioningMode(Enum):
    """Motor positioning mode used in LCN commands."""

    NONE = "NONE"
    BS4 = "BS4"
    MODULE = "MODULE"
    # EMULATED = "EMULATED"


class RelVarRef(Enum):
    """Value-reference for relative LCN variable commands."""

    CURRENT = auto()
    PROG = auto()  # Programmed value (LCN-PRO). Set-points and thresholds.


class SendKeyCommand(Enum):
    """Command types used when sending LCN keys."""

    HIT = "K"
    MAKE = "L"
    BREAK = "O"
    DONTSEND = "-"


class KeyLockStateModifier(Enum):
    """Key-lock modifiers used in LCN commands."""

    ON = "1"
    OFF = "0"
    TOGGLE = "U"
    NOCHANGE = "-"


class ThresholdLockStateModifier(Enum):
    """Threshold-lock modifiers used in LCN commands."""

    ON = "1"
    OFF = "0"
    NOCHANGE = "-"


class BeepSound(Enum):
    """Beep sounds supported by LCN modules."""

    NORMAL = "N"
    SPECIAL = "S"


HARDWARE_DESCRIPTIONS = {
    -1: "UnknownModuleType",
    1: "LCN-SW1.0",
    2: "LCN-SW1.1",
    3: "LCN-UP1.0",
    4: "LCN-UP2",
    5: "LCN-SW2",
    6: "LCN-UP-Profi1-Plus",
    7: "LCN-DI12",
    8: "LCN-HU",
    9: "LCN-SH",
    10: "LCN-UP2",
    11: "LCN-UPP",
    12: "LCN-SK",
    14: "LCN-LD",
    15: "LCN-SH-Plus",
    17: "LCN-UPS",
    18: "LCN_UPS24V",
    19: "LCN-GTM",
    20: "LCN-SHS",
    21: "LCN-ESD",
    22: "LCN-EB2",
    23: "LCN-MRS",
    24: "LCN-EB11",
    25: "LCN-UMR",
    26: "LCN-UPU",
    27: "LCN-UMR24V",
    28: "LCN-SHD",
    29: "LCN-SHU",
    30: "LCN-SR6",
    31: "LCN-UMF",
    32: "LCN-WBH",
}


class HardwareType(Enum):
    """Hardware types as returned by serial number request."""

    UNKNOWN = -1
    SW1_0 = 1
    SW1_1 = 2
    UP1_0 = 3
    UP2 = 4
    SW2 = 5
    UP_PROFI1_PLUS = 6
    DI12 = 7
    HU = 8
    SH = 9
    UPP = 11
    SK = 12
    LD = 14
    SH_PLUS = 15
    UPS = 17
    UPS24V = 18
    GTM = 19
    SHS = 20
    ESD = 21
    EB2 = 22
    MRS = 23
    EB11 = 24
    UMR = 25
    UPU = 26
    UMR24V = 27
    SHD = 28
    SHU = 29
    SR6 = 30
    UMF = 31
    WBH = 32

    @property
    def identifier(self) -> Any:
        """Get the LCN hardware identifier."""
        return self.value

    @property
    def description(self) -> str:
        """Get the LCN hardware name."""
        return HARDWARE_DESCRIPTIONS[self.value]


@no_type_check
def hw_type_new(cls, value):
    """Replace Hardwaretype.__new__."""
    if value == 10:
        value = 4
    return super(HardwareType, cls).__new__(cls, value)


setattr(HardwareType, "__new__", hw_type_new)


class AccessControlPeriphery(Enum):
    """Action types for LCN keys."""

    TRANSMITTER = "transmitter"
    TRANSPONDER = "transponder"
    FINGERPRINT = "fingerprint"
    CODELOCK = "codelock"


class AcknowledgeErrorCode(Enum):
    """Acknowledge error codes."""

    UNKNOWN = -1
    OK = 0
    UNKNOWN_COMMAND = 5
    WRONG_PARAMETER_COUNT = 6
    INVALID_PARAMETER_VALUE = 7
    CURRENTLY_NOT_ALLOWED = 8
    NOT_ALLOWED_BY_PROGRAMMING = 9
    INAPPROPRIATE_MODULE = 10
    MISSING_PERIPHERY = 11
    PROGRAMMING_MODE_REQUIRED = 12
    FUSE_DEFECT = 14

    @classmethod
    def _missing_(cls, value: Any) -> AcknowledgeErrorCode:
        """Handle missing values."""
        return cls.UNKNOWN


class LcnEvent(Enum):
    """LCN events."""

    CONNECTION_ESTABLISHED = "connection-established"
    CONNECTION_LOST = "connection-lost"
    CONNECTION_REFUSED = "connection-refused"
    CONNECTION_TIMEOUT = "connection-timeout"
    PING_TIMEOUT = "ping-timeout"
    TIMEOUT_ERROR = "timeout-error"
    LICENSE_ERROR = "license-error"
    AUTHENTICATION_ERROR = "authentication-error"
    BUS_CONNECTED = "bus-connected"
    BUS_DISCONNECTED = "bus-disconnected"
    BUS_CONNECTION_STATUS_CHANGED = "bus-connection-status-changed"


default_connection_settings: dict[str, Any] = {
    "NUM_TRIES": 3,  # Total number of request to sent before going into
    # failed-state.
    "SK_NUM_TRIES": 3,  # Total number of segment coupler scan tries
    "DIM_MODE": OutputPortDimMode.STEPS50,
    "ACKNOWLEDGE": True,  # modules request an acknowledge command
    "DEFAULT_TIMEOUT": 3.5,  # Default timeout for send command retries
    "MAX_STATUS_EVENTBASED_VALUEAGE": 600,  # Poll interval for
    # status values that
    # automatically send
    # their values on change
    "MAX_STATUS_POLLED_VALUEAGE": 30,  # Poll interval for status
    # values that do not send
    # their values on change
    # (always polled)
    "STATUS_REQUEST_DELAY_AFTER_COMMAND": 2,  # Status request delay
    # after a command has
    # been send which
    # potentially changed
    # that status
    "MAX_RESPONSE_AGE": 60,  # Age in seconds after which stored responses are purged
    "BUS_IDLE_TIME": 0.05,  # Time to wait for message traffic before sending
    "PING_SEND_DELAY": 600,  # The default timeout for pings sent to PCHK
    "PING_RECV_TIMEOUT": 10,  # The default timeout for pings expected from PCHK
    "PING_MODULE_TIMEOUT": 60,  # The delay before sending a ping to a module
}



"""Helper functions for pypck."""

import asyncio
import contextlib
from collections.abc import Coroutine
from typing import Any


async def cancel_task(task: "asyncio.Task[Any]") -> bool:
    """Cancel a task.

    Wait for cancellation completed but do not propagate a possible CancelledError.
    """
    success = task.cancel()
    with contextlib.suppress(asyncio.CancelledError):
        await task

    return success  # was not already done


class TaskRegistry:
    """Keep track of running tasks."""

    def __init__(self) -> None:
        """Init task registry instance."""
        self.tasks: list[asyncio.Task[Any]] = []

    def remove_task(self, task: "asyncio.Task[None]") -> None:
        """Remove a task from the task registry."""
        if task in self.tasks:
            self.tasks.remove(task)

    def create_task(self, coro: Coroutine[Any, Any, Any]) -> "asyncio.Task[None]":
        """Create a task and store a reference in the task registry."""
        task: asyncio.Task[Any] = asyncio.create_task(coro)
        task.add_done_callback(self.remove_task)
        self.tasks.append(task)
        return task

    async def cancel_all_tasks(self) -> None:
        """Cancel all pypck tasks."""
        while self.tasks:
            await cancel_task(self.tasks.pop())



"""Test the data flow for Input objects."""

from unittest.mock import patch

from pypck.inputs import Input, ModInput
from pypck.lcn_addr import LcnAddr

from .conftest import MockPchkConnectionManager


async def test_message_to_input(pypck_client: MockPchkConnectionManager) -> None:
    """Test data flow from message to input."""
    inp = Input()
    message = "dummy_message"
    with patch.object(
        pypck_client, "async_process_input"
    ) as pypck_client_process_input:
        with patch("pypck.inputs.InputParser.parse", return_value=[inp]) as inp_parse:
            await pypck_client.process_message(message)

        inp_parse.assert_called_with(message)
        pypck_client_process_input.assert_awaited_with(inp)


async def test_physical_to_logical_segment_id(
    pypck_client: MockPchkConnectionManager,
) -> None:
    """Test conversion from logical to physical segment id."""
    pypck_client.local_seg_id = 20
    module = pypck_client.get_device_connection(LcnAddr(20, 7, False))
    assert module.is_group is False
    with (
        patch("tests.conftest.MockPchkConnectionManager.is_ready", return_value=True),
        patch.object(module, "async_process_input") as module_process_input,
    ):
        inp = ModInput(LcnAddr(20, 7, False))
        await pypck_client.async_process_input(inp)

        inp = ModInput(LcnAddr(0, 7, False))
        await pypck_client.async_process_input(inp)

        inp = ModInput(LcnAddr(4, 7, False))
        await pypck_client.async_process_input(inp)

        assert module_process_input.await_count == 3



"""Module connection tests."""

import pytest

from pypck.lcn_addr import LcnAddr

from .conftest import MockPchkConnectionManager

TEST_VECTORS = {
    # empty
    "": (b"", b"", b"", b"", b""),
    # pure ascii
    **{"a" * n: (b"a" * n, b"", b"", b"", b"") for n in (1, 7, 11, 12)},
    **{"a" * (12 + n): (b"a" * 12, b"a" * n, b"", b"", b"") for n in (1, 7, 11, 12)},
    **{
        "a" * (48 + n): (b"a" * 12, b"a" * 12, b"a" * 12, b"a" * 12, b"a" * n)
        for n in (1, 7, 11, 12)
    },
    # only two-byte UTF-8
    **{"ü" * n: (b"\xc3\xbc" * n, b"", b"", b"", b"") for n in (1, 5, 6)},
    **{
        "ü" * (6 + n): (b"\xc3\xbc" * 6, b"\xc3\xbc" * n, b"", b"", b"")
        for n in (1, 5, 6)
    },
    **{
        "ü" * (24 + n): (
            b"\xc3\xbc" * 6,
            b"\xc3\xbc" * 6,
            b"\xc3\xbc" * 6,
            b"\xc3\xbc" * 6,
            b"\xc3\xbc" * n,
        )
        for n in (1, 5, 6)
    },
    # only three-byte utf-8
    **{"\u20ac" * n: (b"\xe2\x82\xac" * n, b"", b"", b"", b"") for n in (1, 4)},
    **{
        "\u20ac" * (4 + n): (b"\xe2\x82\xac" * 4, b"\xe2\x82\xac" * n, b"", b"", b"")
        for n in (1, 4)
    },
    **{
        "\u20ac" * (16 + n): (
            b"\xe2\x82\xac" * 4,
            b"\xe2\x82\xac" * 4,
            b"\xe2\x82\xac" * 4,
            b"\xe2\x82\xac" * 4,
            b"\xe2\x82\xac" * n,
        )
        for n in (1, 4)
    },
    # boundary-crossing utf-8
    "12345678123\u00fc4567": (b"12345678123\xc3", b"\xbc4567", b"", b"", b""),
    "12345678123\u20ac4567": (b"12345678123\xe2", b"\x82\xac4567", b"", b"", b""),
    "1234567812\u20ac34567": (b"1234567812\xe2\x82", b"\xac34567", b"", b"", b""),
}


@pytest.mark.parametrize("text, parts", TEST_VECTORS.items())
async def test_dyn_text(
    pypck_client: MockPchkConnectionManager,
    text: str,
    parts: tuple[bytes, bytes, bytes, bytes, bytes],
) -> None:
    """Tests for dynamic text."""
    module = pypck_client.get_device_connection(LcnAddr(0, 10, False))

    await module.dyn_text(3, text)

    module.send_command.assert_awaited()  # type: ignore[attr-defined]
    await_args = (call.args for call in module.send_command.await_args_list)  # type: ignore[attr-defined]
    _, commands = zip(*await_args)

    for i, part in enumerate(parts):
        assert f"GTDT4{i + 1:d}".encode() + part in commands



"""Status requester."""

from __future__ import annotations

import asyncio
import logging
from dataclasses import dataclass, field
from typing import TYPE_CHECKING, Any, TypeVar, cast

from pypck import inputs

if TYPE_CHECKING:
    from pypck.device import DeviceConnection

_LOGGER = logging.getLogger(__name__)

ResponseT = TypeVar("ResponseT", bound=inputs.Input)


@dataclass(unsafe_hash=True)
class StatusRequest[ResponseT]:
    """Data class for status requests."""

    type: type[ResponseT]  # Type of the input expected as response
    parameters: frozenset[tuple[str, Any]]  # {(parameter_name, parameter_value)}
    timestamp: float = field(
        compare=False
    )  # timestamp the response was received; -1=no timestamp
    response: asyncio.Future[ResponseT] = field(
        compare=False
    )  # the response input object


class StatusRequester:
    """Handling of status requests."""

    current_request: StatusRequest[inputs.Input] | None

    def __init__(
        self,
        device_connection: DeviceConnection,
    ) -> None:
        """Initialize the context."""
        self.device_connection = device_connection
        self.current_request = None
        self.request_cache: set[StatusRequest[inputs.Input]] = set()
        self.max_response_age = self.device_connection.conn.settings["MAX_RESPONSE_AGE"]
        self.request_lock = asyncio.Lock()

        self.unregister_inputs = self.device_connection.register_for_inputs(
            self.input_callback
        )

    def get_status_requests(
        self,
        request_type: type[ResponseT],
        parameters: frozenset[tuple[str, Any]] | None = None,
        max_age: int = 0,
    ) -> list[StatusRequest[ResponseT]]:
        """Get the status requests for the given type and parameters."""
        if parameters is None:
            parameters = frozenset()
        results = [
            request
            for request in self.request_cache
            if request.type == request_type
            and parameters.issubset(request.parameters)
            and (
                (request.timestamp == -1)
                or (max_age == -1)
                or (asyncio.get_running_loop().time() - request.timestamp < max_age)
            )
        ]
        results.sort(key=lambda request: request.timestamp, reverse=True)
        return cast(list[StatusRequest[ResponseT]], results)

    def input_callback(self, inp: inputs.Input) -> None:
        """Handle incoming inputs and set the result for the corresponding requests."""
        # Update current request (if it exists)
        if (
            self.current_request is not None
            and not self.current_request.response.done()
        ):
            if isinstance(inp, self.current_request.type) and all(
                getattr(inp, parameter_name) == parameter_value
                for parameter_name, parameter_value in self.current_request.parameters
            ):
                self.current_request.timestamp = asyncio.get_running_loop().time()
                self.current_request.response.set_result(inp)

        # Update cached requests
        for request in self.get_status_requests(type(inp), max_age=-1):
            if all(
                getattr(inp, parameter_name) == parameter_value
                for parameter_name, parameter_value in request.parameters
            ):
                request.timestamp = asyncio.get_running_loop().time()
                request.response = asyncio.get_running_loop().create_future()
                request.response.set_result(inp)

    async def request(
        self,
        response_type: type[ResponseT],
        request_pck: str,
        request_acknowledge: bool = False,
        max_age: int = 0,  # -1: no age limit / infinite age
        **request_kwargs: Any,
    ) -> ResponseT | None:
        """Execute a status request and wait for the response."""
        async with self.request_lock:
            # check for matching request in cache
            if requests := self.get_status_requests(
                response_type,
                frozenset(request_kwargs.items()),
                max_age,
            ):
                _LOGGER.debug(
                    "from %s: %s (cached)",
                    self.device_connection.conn.connection_id,
                    requests[0].response.result().pck,
                )
                return requests[0].response.result()

            # no matching request in cache
            self.current_request = StatusRequest(
                response_type,
                frozenset(request_kwargs.items()),
                -1,
                asyncio.get_running_loop().create_future(),
            )

            result = None
            # send the request up to NUM_TRIES and wait for response future completion
            for _ in range(self.device_connection.conn.settings["NUM_TRIES"]):
                await self.device_connection.send_command(
                    request_acknowledge, request_pck
                )

                try:
                    async with asyncio.timeout(
                        self.device_connection.conn.settings["DEFAULT_TIMEOUT"]
                    ):
                        # Need to shield the future. Otherwise it would get cancelled.
                        result = await asyncio.shield(self.current_request.response)
                        break
                except TimeoutError:
                    continue
                except asyncio.CancelledError:
                    break

            if result is not None:  # add request to cache
                self.request_cache.discard(self.current_request)
                self.request_cache.add(self.current_request)

            return cast(ResponseT | None, result)
